# coding=UTF-8

from natcap.invest.ui import model, inputs
from natcap.invest import wind_energy, MODEL_METADATA


class WindEnergy(model.InVESTModel):
    def __init__(self):
        model.InVESTModel.__init__(
            self,
            label=MODEL_METADATA['wind_energy'].model_title,
            target=wind_energy.execute,
            validator=wind_energy.validate,
            localdoc=MODEL_METADATA['wind_energy'].userguide
        )

        self.wind_data = inputs.File(
            args_key='wind_data_path',
            helptext=(
                "A CSV file that represents the wind input data "
                "(Weibull parameters). Please see the User's Guide for "
                "a more detailed description of the parameters."),
            label='Wind Data Points (CSV)',
            validator=self.validator)
        self.add_input(self.wind_data)
        self.aoi = inputs.File(
            args_key='aoi_vector_path',
            helptext=(
                "Optional.  An OGR-supported vector file containing a "
                "single polygon defining the area of interest.  The "
                "AOI must be projected with linear units equal to "
                "meters.  If the AOI is provided it will clip and "
                "project the outputs to that of the AOI. The Distance "
                "inputs are dependent on the AOI and will only be "
                "accessible if the AOI is selected.  If the AOI is "
                "selected and the Distance parameters are selected, "
                "then the AOI should also cover a portion of the land "
                "polygon to calculate distances correctly.  An AOI is "
                "required for valuation."),
            label='Area Of Interest (Vector) (Optional)',
            validator=self.validator)
        self.add_input(self.aoi)
        self.bathymetry = inputs.File(
            args_key='bathymetry_path',
            helptext=(
                "A GDAL-supported raster file containing elevation "
                "values represented in meters for the area of "
                "interest.  The DEM should cover at least the entire "
                "span of the area of interest and if no AOI is "
                "provided then the default global DEM should be used."),
            label='Bathymetric Digital Elevation Model (Raster)',
            validator=self.validator)
        self.add_input(self.bathymetry)
        self.land_polygon = inputs.File(
            args_key='land_polygon_vector_path',
            helptext=(
                "An OGR-supported polygon vector that represents the "
                "land and coastline that is of interest.  For this "
                "input to be selectable the AOI must be selected.  The "
                "AOI should also cover a portion of this land polygon "
                "to properly calculate distances.  This coastal "
                "polygon, and the area covered by the AOI, form the "
                "basis for distance calculations for wind farm "
                "electrical transmission.  This input is required for "
                "masking by distance values and for valuation."),
            interactive=False,
            label='Land Polygon for Distance Calculation (Vector)',
            validator=self.validator)
        self.add_input(self.land_polygon)
        self.global_wind_parameters = inputs.File(
            args_key='global_wind_parameters_path',
            helptext=(
                "A CSV file that holds wind energy model parameters "
                "for both the biophysical and valuation modules. "
                "These parameters are defaulted to values that are "
                "supported and reviewed in the User's Guide.  It is "
                "recommended that careful consideration be taken "
                "before changing these values and to make a new CSV "
                "file so that the default one always remains."),
            label='Global Wind Energy Parameters (CSV)',
            validator=self.validator)
        self.add_input(self.global_wind_parameters)
        self.turbine_group = inputs.Container(
            label='Turbine Properties')
        self.add_input(self.turbine_group)
        self.turbine_parameters = inputs.File(
            args_key='turbine_parameters_path',
            helptext=(
                "A CSV file that contains parameters corresponding to "
                "a specific turbine type.  The InVEST package comes "
                "with two turbine model options, 3.6 MW and 5.0 MW. A "
                "new turbine class may be created by using the "
                "existing file format conventions and filling in new "
                "parameters.  Likewise an existing class may be "
                "modified according to the user's needs.  It is "
                "recommended that the existing default CSV files are "
                "not overwritten."),
            label='Turbine Type Parameters File (CSV)',
            validator=self.validator)
        self.turbine_group.add_input(self.turbine_parameters)
        self.number_of_machines = inputs.Text(
            args_key='number_of_turbines',
            helptext=(
                "An integer value indicating the number of wind "
                "turbines per wind farm."),
            label='Number Of Turbines',
            validator=self.validator)
        self.turbine_group.add_input(self.number_of_machines)
        self.min_depth = inputs.Text(
            args_key='min_depth',
            helptext=(
                "A floating point value in meters for the minimum "
                "depth of the offshore wind farm installation."),
            label='Minimum Depth for Offshore Wind Farm Installation (meters)',
            validator=self.validator)
        self.turbine_group.add_input(self.min_depth)
        self.max_depth = inputs.Text(
            args_key='max_depth',
            helptext=(
                "A floating point value in meters for the maximum "
                "depth of the offshore wind farm installation."),
            label='Maximum Depth for Offshore Wind Farm Installation (meters)',
            validator=self.validator)
        self.turbine_group.add_input(self.max_depth)
        self.min_distance = inputs.Text(
            args_key='min_distance',
            helptext=(
                "A floating point value in meters that represents the "
                "minimum distance from shore for offshore wind farm "
                "installation.  Required for valuation."),
            interactive=False,
            label=(
                'Minimum Distance for Offshore Wind Farm Installation '
                '(meters)'),
            validator=self.validator)
        self.turbine_group.add_input(self.min_distance)
        self.max_distance = inputs.Text(
            args_key='max_distance',
            helptext=(
                "A floating point value in meters that represents the "
                "maximum distance from shore for offshore wind farm "
                "installation.  Required for valuation."),
            interactive=False,
            label=(
                'Maximum Distance for Offshore Wind Farm Installation '
                '(meters)'),
            validator=self.validator)
        self.turbine_group.add_input(self.max_distance)
        self.valuation_container = inputs.Container(
            args_key='valuation_container',
            expandable=True,
            expanded=False,
            label='Valuation')
        self.add_input(self.valuation_container)
        self.foundation_cost = inputs.Text(
            args_key='foundation_cost',
            helptext=(
                "A floating point number for the unit cost of the "
                "foundation type. The cost of "
                "a foundation will depend on the type selected, which "
                "itself depends on a variety of factors including "
                "depth and turbine choice.  Please see the User's "
                "Guide for guidance on properly selecting this value."),
            label='Cost of the Foundation Type',
            validator=self.validator)
        self.valuation_container.add_input(self.foundation_cost)
        self.discount_rate = inputs.Text(
            args_key='discount_rate',
            helptext=(
                "The discount rate reflects preferences for immediate "
                "benefits over future benefits (e.g., would an "
                "individual rather receive $10 today or $10 five years "
                "from now?). See the User's Guide for guidance on "
                "selecting this value."),
            label='Discount Rate',
            validator=self.validator)
        self.valuation_container.add_input(self.discount_rate)
        self.grid_points = inputs.File(
            args_key='grid_points_path',
            helptext=(
                "An optional CSV file with grid and land points to "
                "determine cable distances from.  An example:<br/> "
                "<table border='1'> <tr> <th>ID</th> <th>TYPE</th> "
                "<th>LATI</th> <th>LONG</th> </tr> <tr> <td>1</td> "
                "<td>GRID</td> <td>42.957</td> <td>-70.786</td> </tr> "
                "<tr> <td>2</td> <td>LAND</td> <td>42.632</td> "
                "<td>-71.143</td> </tr> <tr> <td>3</td> <td>LAND</td> "
                "<td>41.839</td> <td>-70.394</td> </tr> </table> "
                "<br/><br/>Each point location is represented as a "
                "single row with columns being <b>ID</b>, <b>TYPE</b>, "
                "<b>LATI</b>, and <b>LONG</b>. The <b>LATI</b> and "
                "<b>LONG</b> columns indicate the coordinates for the "
                "point.  The <b>TYPE</b> column relates to whether it "
                "is a land or grid point.  The <b>ID</b> column is a "
                "simple unique integer.  The shortest distance between "
                "respective points is used for calculations.  See the "
                "User's Guide for more information."),
            label='Grid Connection Points (Optional)',
            validator=self.validator)
        self.valuation_container.add_input(self.grid_points)
        self.avg_grid_dist = inputs.Text(
            args_key='avg_grid_distance',
            helptext=(
                "<b>Always required, but NOT used in the model if "
                "Grid Points provided</b><br/><br/>A number in "
                "kilometres that is only used if grid points are NOT "
                "used in valuation.  When running valuation using the "
                "land polygon to compute distances, the model uses an "
                "average distance to the onshore grid from coastal "
                "cable landing points instead of specific grid "
                "connection points.  See the User's Guide for a "
                "description of the approach and the method used to "
                "calculate the default value."),
            label='Average Shore to Grid Distance (Kilometers)',
            validator=self.validator)
        self.valuation_container.add_input(self.avg_grid_dist)
        self.price_table = inputs.Checkbox(
            args_key='price_table',
            helptext=(
                "When checked the model will use the social cost of "
                "wind energy table provided in the input below.  If "
                "not checked the price per year will be determined "
                "using the price of energy input and the annual rate "
                "of change."),
            label='Use Price Table')
        self.valuation_container.add_input(self.price_table)
        self.wind_schedule = inputs.File(
            args_key='wind_schedule',
            helptext=(
                "A CSV file that has the price of wind energy per "
                "kilowatt hour for each year of the wind farms life. "
                "The CSV file should have the following two "
                "columns:<br/><br/><b>Year:</b> a set of integers "
                "indicating each year for the lifespan of the wind "
                "farm.  They can be in date form such as : 2010, 2011, "
                "2012... OR simple time step integers such as : 0, 1, "
                "2... <br/><br/><b>Price:</b> a set of floats "
                "indicating the price of wind energy per kilowatt hour "
                "for a particular year or time step in the wind farms "
                "life.<br/><br/>An example:<br/> <table border='1'> "
                "<tr><th>Year</th> <th>Price</th></tr><tr><td>0</td><t "
                "d>.244</td></tr><tr><td>1</td><td>.255</td></tr><tr>< "
                "td>2</td><td>.270</td></tr><tr><td>3</td><td>.275</td "
                "></tr><tr><td>4</td><td>.283</td></tr><tr><td>5</td>< "
                "td>.290</td></tr></table><br/><br/><b>NOTE:</b> The "
                "number of years or time steps listed must match the "
                "<b>time</b> parameter in the <b>Global Wind Energy "
                "Parameters</b> input file above.  In the above "
                "example we have 6 years for the lifetime of the farm, "
                "year 0 being a construction year and year 5 being the "
                "last year."),
            interactive=False,
            label='Wind Energy Price Table (CSV)',
            validator=self.validator)
        self.valuation_container.add_input(self.wind_schedule)
        self.wind_price = inputs.Text(
            args_key='wind_price',
            helptext=(
                "The price of energy per kilowatt hour.  This is the "
                "price that will be used for year or time step 0 and "
                "will then be adjusted based on the rate of change "
                "percentage from the input below.  See the User's "
                "Guide for guidance about determining this value."),
            label='Price of Energy per Kilowatt Hour (currency/kWh)',
            validator=self.validator)
        self.valuation_container.add_input(self.wind_price)
        self.rate_change = inputs.Text(
            args_key='rate_change',
            helptext=(
                "The annual rate of change in the price of wind "
                "energy.  This should be expressed as a decimal "
                "percentage.  For example, 0.1 for a 10% annual price "
                "change."),
            label='Annual Rate of Change in Price of Wind Energy',
            validator=self.validator)
        self.valuation_container.add_input(self.rate_change)

        # Set interactivity, requirement as input sufficiency changes
        self.aoi.sufficiency_changed.connect(
            self.land_polygon.set_interactive)
        self.land_polygon.sufficiency_changed.connect(
            self.min_distance.set_interactive)
        self.land_polygon.sufficiency_changed.connect(
            self.max_distance.set_interactive)
        self.price_table.sufficiency_changed.connect(
            self._toggle_price_options)

    def _toggle_price_options(self, use_price_table):
        self.wind_schedule.set_interactive(use_price_table)
        self.wind_price.set_interactive(not use_price_table)
        self.rate_change.set_interactive(not use_price_table)

    def assemble_args(self):
        args = {
            self.workspace.args_key: self.workspace.value(),
            self.suffix.args_key: self.suffix.value(),
            self.wind_data.args_key: self.wind_data.value(),
            self.bathymetry.args_key: self.bathymetry.value(),
            self.global_wind_parameters.args_key:
                self.global_wind_parameters.value(),
            self.turbine_parameters.args_key: self.turbine_parameters.value(),
            self.number_of_machines.args_key: self.number_of_machines.value(),
            self.min_depth.args_key: self.min_depth.value(),
            self.max_depth.args_key: self.max_depth.value(),
            self.valuation_container.args_key: self.valuation_container.value(),
            self.avg_grid_dist.args_key: self.avg_grid_dist.value(),
        }
        if self.aoi.value():
            args[self.aoi.args_key] = self.aoi.value()
        if self.land_polygon.value():
            args[self.land_polygon.args_key] = self.land_polygon.value()
        if self.min_distance.value():
            args[self.min_distance.args_key] = self.min_distance.value()
        if self.max_distance.value():
            args[self.max_distance.args_key] = self.max_distance.value()
        if self.grid_points.value():
            args[self.grid_points.args_key] = self.grid_points.value()

        # Include these args if valuation is checked.
        if args[self.valuation_container.args_key]:
            args[self.foundation_cost.args_key] = self.foundation_cost.value()
            args[self.discount_rate.args_key] = self.discount_rate.value()
            args[self.price_table.args_key] = self.price_table.value()
            args[self.wind_schedule.args_key] = self.wind_schedule.value()
            args[self.wind_price.args_key] = self.wind_price.value()
            args[self.rate_change.args_key] = self.rate_change.value()

        return args



"""init module for natcap.invest."""
import builtins
import dataclasses
import gettext
import logging
import os
import sys

import babel
import pkg_resources

LOGGER = logging.getLogger('natcap.invest')
LOGGER.addHandler(logging.NullHandler())
__all__ = ['local_dir', ]

# location of our translation message catalog directory
LOCALE_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'internationalization/locales')

# all supported language codes, including the default English
LOCALES = sorted(set(os.listdir(LOCALE_DIR) + ['en']))

# map locale codes to the corresponding localized language name
# e.g. 'es': 'español'
LOCALE_NAME_MAP = {
    locale: babel.Locale(locale).display_name for locale in LOCALES
}

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    # package is not installed.  Log the exception for debugging.
    LOGGER.exception('Could not load natcap.invest version information')

# Check if the function _() is available
# If not, define it as the identity function
# _() is installed into builtins by gettext when we set up to translate
# It wraps every string in every model that we want to translate
# Make sure it's defined so that natcap.invest modules are importable whether
# or not gettext has been installed in the importing namespace
if not callable(getattr(builtins, '_', None)):
    def identity(x): return x
    builtins.__dict__['_'] = identity


@dataclasses.dataclass
class _MODELMETA:
    """Dataclass to store frequently used model metadata."""
    model_title: str  # display name for the model
    pyname: str       # importable python module name for the model
    gui: str          # importable python class for the corresponding Qt UI
    userguide: str    # name of the corresponding built userguide file
    aliases: tuple    # alternate names for the model, if any


MODEL_METADATA = {
    'annual_water_yield': _MODELMETA(
        model_title=_('Annual Water Yield'),
        pyname='natcap.invest.annual_water_yield',
        gui='annual_water_yield.AnnualWaterYield',
        userguide='annual_water_yield.html',
        aliases=('hwy', 'awy')),
    'carbon': _MODELMETA(
        model_title=_('Carbon Storage and Sequestration'),
        pyname='natcap.invest.carbon',
        gui='carbon.Carbon',
        userguide='carbonstorage.html',
        aliases=()),
    'coastal_blue_carbon': _MODELMETA(
        model_title=_('Coastal Blue Carbon'),
        pyname='natcap.invest.coastal_blue_carbon.coastal_blue_carbon',
        gui='cbc.CoastalBlueCarbon',
        userguide='coastal_blue_carbon.html',
        aliases=('cbc',)),
    'coastal_blue_carbon_preprocessor': _MODELMETA(
        model_title=_('Coastal Blue Carbon Preprocessor'),
        pyname='natcap.invest.coastal_blue_carbon.preprocessor',
        gui='cbc.CoastalBlueCarbonPreprocessor',
        userguide='coastal_blue_carbon.html',
        aliases=('cbc_pre',)),
    'coastal_vulnerability': _MODELMETA(
        model_title=_('Coastal Vulnerability'),
        pyname='natcap.invest.coastal_vulnerability',
        gui='coastal_vulnerability.CoastalVulnerability',
        userguide='coastal_vulnerability.html',
        aliases=('cv',)),
    'crop_production_percentile': _MODELMETA(
        model_title=_('Crop Production: Percentile'),
        pyname='natcap.invest.crop_production_percentile',
        gui='crop_production.CropProductionPercentile',
        userguide='crop_production.html',
        aliases=('cpp',)),
    'crop_production_regression': _MODELMETA(
        model_title=_('Crop Production: Regression'),
        pyname='natcap.invest.crop_production_regression',
        gui='crop_production.CropProductionRegression',
        userguide='crop_production.html',
        aliases=('cpr',)),
    'delineateit': _MODELMETA(
        model_title=_('DelineateIt'),
        pyname='natcap.invest.delineateit.delineateit',
        gui='delineateit.Delineateit',
        userguide='delineateit.html',
        aliases=()),
    'forest_carbon_edge_effect': _MODELMETA(
        model_title=_('Forest Carbon Edge Effect'),
        pyname='natcap.invest.forest_carbon_edge_effect',
        gui='forest_carbon.ForestCarbonEdgeEffect',
        userguide='carbon_edge.html',
        aliases=('fc',)),
    'globio': _MODELMETA(
        model_title=_('GLOBIO'),
        pyname='natcap.invest.globio',
        gui='globio.GLOBIO',
        userguide='globio.html',
        aliases=()),
    'habitat_quality': _MODELMETA(
        model_title=_('Habitat Quality'),
        pyname='natcap.invest.habitat_quality',
        gui='habitat_quality.HabitatQuality',
        userguide='habitat_quality.html',
        aliases=('hq',)),
    'habitat_risk_assessment': _MODELMETA(
        model_title=_('Habitat Risk Assessment'),
        pyname='natcap.invest.hra',
        gui='hra.HabitatRiskAssessment',
        userguide='habitat_risk_assessment.html',
        aliases=('hra',)),
    'ndr': _MODELMETA(
        model_title=_('Nutrient Delivery Ratio'),
        pyname='natcap.invest.ndr.ndr',
        gui='ndr.Nutrient',
        userguide='ndr.html',
        aliases=()),
    'pollination': _MODELMETA(
        model_title=_('Crop Pollination'),
        pyname='natcap.invest.pollination',
        gui='pollination.Pollination',
        userguide='croppollination.html',
        aliases=()),
    'recreation': _MODELMETA(
        model_title=_('Visitation: Recreation and Tourism'),
        pyname='natcap.invest.recreation.recmodel_client',
        gui='recreation.Recreation',
        userguide='recreation.html',
        aliases=()),
    'routedem': _MODELMETA(
        model_title=_('RouteDEM'),
        pyname='natcap.invest.routedem',
        gui='routedem.RouteDEM',
        userguide='routedem.html',
        aliases=()),
    'scenario_generator_proximity': _MODELMETA(
        model_title=_('Scenario Generator: Proximity Based'),
        pyname='natcap.invest.scenario_gen_proximity',
        gui='scenario_gen.ScenarioGenProximity',
        userguide='scenario_gen_proximity.html',
        aliases=('sgp',)),
    'scenic_quality': _MODELMETA(
        model_title=_('Scenic Quality'),
        pyname='natcap.invest.scenic_quality.scenic_quality',
        gui='scenic_quality.ScenicQuality',
        userguide='scenic_quality.html',
        aliases=('sq',)),
    'sdr': _MODELMETA(
        model_title=_('Sediment Delivery Ratio'),
        pyname='natcap.invest.sdr.sdr',
        gui='sdr.SDR',
        userguide='sdr.html',
        aliases=()),
    'seasonal_water_yield': _MODELMETA(
        model_title=_('Seasonal Water Yield'),
        pyname='natcap.invest.seasonal_water_yield.seasonal_water_yield',
        gui='seasonal_water_yield.SeasonalWaterYield',
        userguide='seasonal_water_yield.html',
        aliases=('swy',)),
    'stormwater': _MODELMETA(
        model_title=_('Urban Stormwater Retention'),
        pyname='natcap.invest.stormwater',
        gui='stormwater.Stormwater',
        userguide='stormwater.html',
        aliases=()),
    'wave_energy': _MODELMETA(
        model_title=_('Wave Energy Production'),
        pyname='natcap.invest.wave_energy',
        gui='wave_energy.WaveEnergy',
        userguide='wave_energy.html',
        aliases=()),
    'wind_energy': _MODELMETA(
        model_title=_('Wind Energy Production'),
        pyname='natcap.invest.wind_energy',
        gui='wind_energy.WindEnergy',
        userguide='wind_energy.html',
        aliases=()),
    'urban_flood_risk_mitigation': _MODELMETA(
        model_title=_('Urban Flood Risk Mitigation'),
        pyname='natcap.invest.urban_flood_risk_mitigation',
        gui='urban_flood_risk_mitigation.UrbanFloodRiskMitigation',
        userguide='urban_flood_mitigation.html',
        aliases=('ufrm',)),
    'urban_cooling_model': _MODELMETA(
        model_title=_('Urban Cooling'),
        pyname='natcap.invest.urban_cooling_model',
        gui='urban_cooling_model.UrbanCoolingModel',
        userguide='urban_cooling_model.html',
        aliases=('ucm',)),
}


def install_language(language_code):
    """Globally install the _() function for the requested language.

    Args:
        language_code (str): ISO 639-1 locale code for a language supported
            by invest

    Returns:
        None
    """
    if language_code not in LOCALES:
        raise ValueError(
            f"Language '{language_code}' is not supported by InVEST. "
            f"Supported language codes are: {LOCALES}")
    language = gettext.translation(
        'messages',
        languages=[language_code],
        localedir=LOCALE_DIR,
        # fall back to a NullTranslation, which returns the English messages
        fallback=True)
    language.install()
    LOGGER.debug(f'Installed language "{language_code}"')


def local_dir(source_file):
    """Return the path to where `source_file` would be on disk.

    If this is frozen (as with PyInstaller), this will be the folder with the
    executable in it.  If not, it'll just be the foldername of the source_file
    being passed in.
    """
    source_dirname = os.path.dirname(source_file)
    if getattr(sys, 'frozen', False):
        # sys.frozen is True when we're in either a py2exe or pyinstaller
        # build.
        # sys._MEIPASS exists, we're in a Pyinstaller build.
        if not getattr(sys, '_MEIPASS', False):
            # only one os.path.dirname() results in the path being relative to
            # the natcap.invest package, when I actually want natcap/invest to
            # be in the filepath.

            # relpath would be something like <modelname>/<data_file>
            relpath = os.path.relpath(source_file, os.path.dirname(__file__))
            pkg_path = os.path.join('natcap', 'invest', relpath)
            return os.path.join(
                os.path.dirname(sys.executable), os.path.dirname(pkg_path))
        else:
            # assume that if we're in a frozen build, we're in py2exe.  When in
            # py2exe, the directory structure is maintained, so we just return
            # the source_dirname.
            pass
    return source_dirname



# coding=UTF-8

import functools

from natcap.invest.ui import model, inputs
from natcap.invest.coastal_blue_carbon import coastal_blue_carbon, preprocessor
from natcap.invest import MODEL_METADATA


def _create_input_kwargs_from_args_spec(
        args_key, args_spec, validator):
    """Helper function to return kwargs for most model inputs.

    Args:
        args_key: The args key of the input from which a kwargs
            dict is being built.
        args_spec: The ARGS_SPEC object to reference.
        validator: The validator callable to provide to the ``validator`` kwarg
            for the input.

    Returns:
        A dict of ``kwargs`` to explode to an ``inputs.GriddedInput``
        object at creation time.
    """
    model_spec = args_spec['args']
    return {
        'args_key': args_key,
        'helptext': model_spec[args_key]['about'],
        'label': model_spec[args_key]['name'],
        'validator': validator,
    }


class CoastalBlueCarbonPreprocessor(model.InVESTModel):
    def __init__(self):
        model.InVESTModel.__init__(
            self,
            label=MODEL_METADATA['coastal_blue_carbon_preprocessor'].model_title,
            target=preprocessor.execute,
            validator=preprocessor.validate,
            localdoc=MODEL_METADATA['coastal_blue_carbon_preprocessor'].userguide)

        _ui_keys = functools.partial(
            _create_input_kwargs_from_args_spec,
            args_spec=preprocessor.ARGS_SPEC,
            validator=self.validator)

        self.lulc_snapshot_csv = inputs.File(
            **_ui_keys('landcover_snapshot_csv'))
        self.add_input(self.lulc_snapshot_csv)

        self.lulc_lookup_table_path = inputs.File(
            **_ui_keys('lulc_lookup_table_path'))
        self.add_input(self.lulc_lookup_table_path)

    def assemble_args(self):
        args = {
            self.workspace.args_key: self.workspace.value(),
            self.suffix.args_key: self.suffix.value(),
            self.lulc_lookup_table_path.args_key:
                self.lulc_lookup_table_path.value(),
            self.lulc_snapshot_csv.args_key: self.lulc_snapshot_csv.value(),
        }
        return args


class CoastalBlueCarbon(model.InVESTModel):
    def __init__(self):
        model.InVESTModel.__init__(
            self,
            label=MODEL_METADATA['coastal_blue_carbon'].model_title,
            target=coastal_blue_carbon.execute,
            validator=coastal_blue_carbon.validate,
            localdoc=MODEL_METADATA['coastal_blue_carbon'].userguide)

        _ui_keys = functools.partial(
            _create_input_kwargs_from_args_spec,
            args_spec=coastal_blue_carbon.ARGS_SPEC,
            validator=self.validator)

        self.snapshots_table = inputs.File(
            **_ui_keys('landcover_snapshot_csv'))
        self.add_input(self.snapshots_table)

        self.biophysical_table_path = inputs.File(
            **_ui_keys('biophysical_table_path'))
        self.add_input(self.biophysical_table_path)

        self.landcover_transitions_table = inputs.File(
            **_ui_keys('landcover_transitions_table'))
        self.add_input(self.landcover_transitions_table)

        self.analysis_year = inputs.Text(**_ui_keys('analysis_year'))
        self.add_input(self.analysis_year)

        self.do_economic_analysis = inputs.Container(
            args_key='do_economic_analysis',
            expandable=True,
            expanded=True,
            label='Calculate Net Present Value of Sequestered Carbon')
        self.add_input(self.do_economic_analysis)

        self.use_price_table = inputs.Checkbox(
            args_key='use_price_table',
            helptext='',
            label='Use Price Table')
        self.do_economic_analysis.add_input(self.use_price_table)
        self.price = inputs.Text(**_ui_keys('price'))
        self.do_economic_analysis.add_input(self.price)

        self.inflation_rate = inputs.Text(**_ui_keys('inflation_rate'))
        self.do_economic_analysis.add_input(self.inflation_rate)

        self.price_table_path = inputs.File(**_ui_keys('price_table_path'))
        self.do_economic_analysis.add_input(self.price_table_path)

        self.discount_rate = inputs.Text(**_ui_keys('discount_rate'))
        self.do_economic_analysis.add_input(self.discount_rate)

        # Set interactivity, requirement as input sufficiency changes
        self.use_price_table.sufficiency_changed.connect(
            self._price_table_sufficiency_changed)

    def _price_table_sufficiency_changed(self, new_sufficiency):
        self.price.set_interactive(not new_sufficiency)
        self.inflation_rate.set_interactive(not new_sufficiency)
        self.price_table_path.set_interactive(new_sufficiency)

    def assemble_args(self):
        args = {
            self.workspace.args_key: self.workspace.value(),
            self.suffix.args_key: self.suffix.value(),
            self.snapshots_table.args_key:
                self.snapshots_table.value(),
            self.landcover_transitions_table.args_key:
                self.landcover_transitions_table.value(),
            self.analysis_year.args_key: self.analysis_year.value(),
            self.biophysical_table_path.args_key:
                self.biophysical_table_path.value(),
            self.landcover_transitions_table.args_key:
                self.landcover_transitions_table.value(),
            self.do_economic_analysis.args_key:
                self.do_economic_analysis.value(),
            }

        if self.do_economic_analysis.value():
            args[self.price.args_key] = self.price.value()
            args[self.inflation_rate.args_key] = self.inflation_rate.value()
            args[self.discount_rate.args_key] = self.discount_rate.value()

            args[self.use_price_table.args_key] = self.use_price_table.value()
            if self.use_price_table.value():
                args[self.price_table_path.args_key] = (
                    self.price_table_path.value())

        return args



# coding=UTF-8

from natcap.invest.ui import model, inputs
from natcap.invest import habitat_quality, MODEL_METADATA


class HabitatQuality(model.InVESTModel):
    def __init__(self):
        model.InVESTModel.__init__(
            self,
            label=MODEL_METADATA['habitat_quality'].model_title,
            target=habitat_quality.execute,
            validator=habitat_quality.validate,
            localdoc=MODEL_METADATA['habitat_quality'].userguide)
        self.current_landcover = inputs.File(
            args_key='lulc_cur_path',
            helptext=(
                "A GDAL-supported raster file.  The current LULC must "
                "have its' own threat rasters, where each threat "
                "raster file path is defined in the <b>Threats Data</b> "
                "CSV..<br/><br/> "
                "Each cell should represent a LULC code as an Integer. "
                "The dataset should be in a projection where the units "
                "are in meters and the projection used should be "
                "defined.  <b>The LULC codes must match the codes in "
                "the Sensitivity table</b>."),
            label='Current Land Cover (Raster)',
            validator=self.validator)
        self.add_input(self.current_landcover)
        self.future_landcover = inputs.File(
            args_key='lulc_fut_path',
            helptext=(
                "Optional.  A GDAL-supported raster file.  Inputting "
                "a future LULC will generate degradation, habitat "
                "quality, and habitat rarity (If baseline is input) "
                "outputs.  The future LULC must have it's own threat "
                "rasters, where each threat raster file path is defined "
                "in the <b>Threats Data</b> CSV.<br/><br/>Each cell should "
                "represent a LULC code as an Integer.  The dataset "
                "should be in a projection where the units are in "
                "meters and the projection used should be defined. "
                "<b>The LULC codes must match the codes in the "
                "Sensitivity table</b>."),
            label='Future Land Cover (Raster) (Optional)',
            validator=self.validator)
        self.add_input(self.future_landcover)
        self.baseline_landcover = inputs.File(
            args_key='lulc_bas_path',
            helptext=(
                "Optional.  A GDAL-supported raster file.  If the "
                "baseline LULC is provided, rarity outputs will be "
                "created for the current and future LULC. The baseline "
                "LULC can have it's own threat rasters (optional), "
                "where each threat raster file path is defined in the "
                "<b>Threats Data</b> CSV. If there are no threat rasters and "
                "the threat paths are left blank in the CSV column, "
                "degradation and habitat quality outputs will not be "
                "generated for the baseline LULC.<br/><br/> Each cell "
                "should  represent a LULC code as an Integer.  The "
                "dataset should be in a projection where the units are "
                "in meters and the projection used should be defined. "
                "The LULC codes must match the codes in the "
                "Sensitivity table.  If possible the baseline map "
                "should refer to a time when intensive management of "
                "the landscape was relatively rare."),
            label='Baseline Land Cover (Raster) (Optional)',
            validator=self.validator)
        self.add_input(self.baseline_landcover)
        self.threats_data = inputs.File(
            args_key='threats_table_path',
            helptext=(
                "A CSV file of all the threats for the model to consider. "
                "Each row in the table is a degradation source. The columns "
                "(THREAT, MAX_DIST, WEIGHT, DECAY) are different attributes "
                "of each degradation source. The columns (BASE_PATH, CUR_PATH, "
                "FUT_PATH) specify the filepath name for the degradation "
                "source where the path is relative to the THREAT CSV. "
                "Column names are case-insensitive. "
                "<br/><br/><b>THREAT:</b> The name of the "
                "threat source must match exactly to the name "
                "of it's corresponding column in the sensitivity table. "
                "<br/><br/><b>MAX_DIST:</b> A number in kilometres "
                "(km) for the maximum distance a threat has an "
                "affect.<br/><br/><b>WEIGHT:</b> A floating point "
                "value between 0 and 1 for the the threats weight "
                "relative to the other threats.  Depending on the type "
                "of habitat under review, certain threats may cause "
                "greater degradation than other "
                "threats.<br/><br/><b>DECAY:</b> A string value of "
                "either <b>exponential</b> or <b>linear</b> "
                "representing the type of decay over space for the "
                "threat. <br/><br/><b>BASE_PATH:</b> Required if baseline "
                "LULC input. The THREAT raster filepath for the base scenario "
                "where the filepath is relative to the THREAT CSV input. "
                "Entries can be left empty if there is no baseline scenario "
                "or if using the baseline LULC for rarity calculations only. "
                "<br/><br/><b>CUR_PATH:</b> "
                "The THREAT raster filepath for the current scenario "
                "where the filepath is relative to the THREAT CSV input. "
                "<br/><br/><b>FUT_PATH:</b> Required if future LULC input. "
                "The THREAT raster filepath for the future scenario where the "
                "filepath is relative to the THREAT CSV input."
                "<br/><br/>See the user's guide for valid values for these "
                "columns."),
            label='Threats Data',
            validator=self.validator)
        self.add_input(self.threats_data)
        self.accessibility_threats = inputs.File(
            args_key='access_vector_path',
            helptext=(
                "An OGR-supported vector file.  The input contains "
                "data on the relative protection that legal / "
                "institutional / social / physical barriers provide "
                "against threats.  The vector file should contain "
                "polygons with a field <b>ACCESS</b>. The "
                "<b>ACCESS</b> values should range from 0 - 1, where 1 "
                "is fully accessible.  Any cells not covered by a "
                "polygon will be set to 1."),
            label='Accessibility to Threats (Vector) (Optional)',
            validator=self.validator)
        self.add_input(self.accessibility_threats)
        self.sensitivity_data = inputs.File(
            args_key='sensitivity_table_path',
            helptext=(
                "A CSV file of LULC types, whether or not the are "
                "considered habitat, and, for LULC types that are "
                "habitat, their specific sensitivity to each threat. "
                "Each row is a LULC type with the following columns: "
                "<b>LULC, HABITAT, THREAT1, THREAT2, "
                "...</b><br/><br/>. Column names are case-insensitive. "
                "<b>LULC:</b> Integer values that "
                "reflect each LULC code found in current, future, and "
                "baseline rasters.<br/><br/><b>HABITAT:</b> A value of "
                "0 or 1 (presence / absence) or a value between 0 and "
                "1 (continuum) depicting the suitability of "
                "habitat.<br/><br/><b>THREATN:</b> Each THREATN "
                "should match exactly with the threat names given in "
                "the threat CSV file, where the THREATN is the name "
                "that matches.  This is an floating point value "
                "between 0 and 1 that represents the sensitivity of a "
                "habitat to a threat.  <br/><br/>.Please see the users "
                "guide for more detailed information on proper column "
                "values and column names for each threat."),
            label='Sensitivity of Land Cover Types to Each Threat, File (CSV)',
            validator=self.validator)
        self.add_input(self.sensitivity_data)
        self.half_saturation_constant = inputs.Text(
            args_key='half_saturation_constant',
            helptext=(
                "A positive floating point value that is defaulted at "
                "0.05. This is the value of the parameter k in equation "
                "(4). In general, set k to half of the highest grid "
                "cell degradation value on the landscape.  To perform "
                "this model calibration the model must be run once in "
                "order to find the highest degradation value and set k "
                "for the provided landscape.  Note that the choice of "
                "k only determines the spread and central tendency of "
                "habitat quality cores and does not affect the rank."),
            label='Half-Saturation Constant',
            validator=self.validator)
        self.add_input(self.half_saturation_constant)

    def assemble_args(self):
        args = {
            self.workspace.args_key: self.workspace.value(),
            self.suffix.args_key: self.suffix.value(),
            self.current_landcover.args_key: self.current_landcover.value(),
            self.threats_data.args_key: self.threats_data.value(),
            self.sensitivity_data.args_key: self.sensitivity_data.value(),
            self.half_saturation_constant.args_key:
                self.half_saturation_constant.value(),
        }
        if self.future_landcover.value():
            args[self.future_landcover.args_key] = self.future_landcover.value()
        if self.baseline_landcover.value():
            args[self.baseline_landcover.args_key] = (
                self.baseline_landcover.value())
        if self.accessibility_threats.value():
            args[self.accessibility_threats.args_key] = (
                self.accessibility_threats.value())

        return args



# -*- coding: utf-8 -*-
"""Coastal Blue Carbon Preprocessor."""
import time
import os
import logging

from osgeo import gdal
import pygeoprocessing
import taskgraph

from .. import utils
from .. import spec_utils
from ..spec_utils import u
from .. import validation
from .. import MODEL_METADATA
from . import coastal_blue_carbon


LOGGER = logging.getLogger(__name__)
ARGS_SPEC = {
    "model_name": MODEL_METADATA["coastal_blue_carbon_preprocessor"].model_title,
    "pyname": MODEL_METADATA["coastal_blue_carbon_preprocessor"].pyname,
    "userguide_html": MODEL_METADATA["coastal_blue_carbon_preprocessor"].userguide,
    "args": {
        "workspace_dir": spec_utils.WORKSPACE,
        "results_suffix": spec_utils.SUFFIX,
        "n_workers": spec_utils.N_WORKERS,
        "lulc_lookup_table_path": {
            "name": _("LULC lookup table"),
            "type": "csv",
            "about": _(
                "A table mapping LULC codes from the snapshot rasters to the "
                "corresponding LULC class names, and whether or not the "
                "class is a coastal blue carbon habitat."),
            "columns": {
                "code": {
                    "type": "integer",
                    "about": _(
                        "LULC code. Every value in the "
                        "snapshot LULC maps must have a corresponding entry "
                        "in this column.")},
                "lulc-class": {
                    "type": "freestyle_string",
                    "about": _("Name of the LULC class.")},
                "is_coastal_blue_carbon_habitat": {
                    "type": "boolean",
                    "about": _(
                        "Enter TRUE if this LULC class is a coastal blue "
                        "carbon habitat, FALSE if not.")}
            }
        },
        "landcover_snapshot_csv": {
            "type": "csv",
            "columns": {
                "snapshot_year": {
                    "type": "number",
                    "units": u.year,
                    "about": _("Year to snapshot.")},
                "raster_path": {
                    "type": "raster",
                    "bands": {1: {"type": "integer"}},
                    "about": _(
                        "Map of LULC in the snapshot year. "
                        "All values in this raster must have corresponding "
                        "entries in the LULC Lookup table.")
                }
            },
            "about": _(
                "A table mapping snapshot years to corresponding LULC maps "
                "for each year."),
            "name": _("LULC snapshots table"),
        },
    }
}


ALIGNED_LULC_RASTER_TEMPLATE = 'aligned_lulc_{year}{suffix}.tif'
TRANSITION_TABLE = 'carbon_pool_transition_template{suffix}.csv'
BIOPHYSICAL_TABLE = 'carbon_biophysical_table_template{suffix}.csv'


def execute(args):
    """Coastal Blue Carbon Preprocessor.

    The preprocessor accepts a list of rasters and checks for cell-transitions
    across the rasters.  The preprocessor outputs a CSV file representing a
    matrix of land cover transitions, each cell pre-filled with a string
    indicating whether carbon accumulates or is disturbed as a result of the
    transition, if a transition occurs.

    Args:
        args['workspace_dir'] (string): directory path to workspace
        args['results_suffix'] (string): append to outputs directory name if
            provided
        args['lulc_lookup_table_path'] (string): filepath of lulc lookup table
        args['landcover_csv_path'] (string): filepath to a CSV containing the
            year and filepath to snapshot rasters on disk.  The years may be in
            any order, but must be unique.

    Returns:
        ``None``
    """
    suffix = utils.make_suffix_string(args, 'results_suffix')
    output_dir = os.path.join(args['workspace_dir'], 'outputs_preprocessor')
    taskgraph_cache_dir = os.path.join(args['workspace_dir'], 'task_cache')
    utils.make_directories([output_dir, taskgraph_cache_dir])

    try:
        n_workers = int(args['n_workers'])
    except (KeyError, ValueError, TypeError):
        # KeyError when n_workers is not present in args
        # ValueError when n_workers is an empty string.
        # TypeError when n_workers is None.
        n_workers = -1  # Synchronous mode.
    task_graph = taskgraph.TaskGraph(
        taskgraph_cache_dir, n_workers, reporting_interval=5.0)

    snapshots_dict = (
        coastal_blue_carbon._extract_snapshots_from_table(
            args['landcover_snapshot_csv']))

    # Align the raster stack for analyzing the various transitions.
    min_pixel_size = float('inf')
    source_snapshot_paths = []
    aligned_snapshot_paths = []
    for snapshot_year, raster_path in snapshots_dict.items():
        source_snapshot_paths.append(raster_path)
        aligned_snapshot_paths.append(os.path.join(
            output_dir, ALIGNED_LULC_RASTER_TEMPLATE.format(
                year=snapshot_year, suffix=suffix)))
        min_pixel_size = min(
            utils.mean_pixel_size_and_area(
                pygeoprocessing.get_raster_info(raster_path)['pixel_size'])[0],
            min_pixel_size)

    baseline_srs_wkt = pygeoprocessing.get_raster_info(
        snapshots_dict[min(snapshots_dict.keys())])['projection_wkt']
    alignment_task = task_graph.add_task(
        func=pygeoprocessing.align_and_resize_raster_stack,
        args=(source_snapshot_paths,
              aligned_snapshot_paths,
              (['near']*len(source_snapshot_paths)),
              (min_pixel_size, -min_pixel_size),
              'intersection'),
        kwargs={'target_projection_wkt': baseline_srs_wkt},
        target_path_list=aligned_snapshot_paths,
        task_name='Align input landcover rasters')

    landcover_table = utils.build_lookup_from_csv(
        args['lulc_lookup_table_path'], 'code')

    target_transition_table = os.path.join(
        output_dir, TRANSITION_TABLE.format(suffix=suffix))
    _ = task_graph.add_task(
        func=_create_transition_table,
        args=(landcover_table,
              sorted(snapshots_dict.values(), key=lambda x: [0]),
              target_transition_table),
        target_path_list=[target_transition_table],
        dependent_task_list=[alignment_task],
        task_name='Determine transitions and write transition table')

    target_biophysical_table_path = os.path.join(
        output_dir, BIOPHYSICAL_TABLE.format(suffix=suffix))
    _ = task_graph.add_task(
        func=_create_biophysical_table,
        args=(landcover_table, target_biophysical_table_path),
        target_path_list=[target_biophysical_table_path],
        task_name='Write biophysical table template')

    task_graph.close()
    task_graph.join()


def _create_transition_table(landcover_table, lulc_snapshot_list,
                             target_table_path):
    """Create the transition table from a series of landcover snapshots.

    Args:
        landcover_table (dict): A dict mapping integer landcover codes to dict
            values indicating the landcover class name in the ``lulc-class``
            field and ``True`` or ``False`` under the
            ``is_coastal_blue_carbon_habitat`` key.
        lulc_snapshot_list (list): A list of string paths to GDAL rasters on
            disk.  All rasters must have the same spatial reference, pixel size
            and dimensions and must also all be integer rasters, where all
            non-nodata pixel values must be represented in the
            ``landcover_table`` dict.
        target_table_path (string): A string path to where the target
            transition table should be written.

    Returns:
        ``None``.
    """
    n_rows, n_cols = pygeoprocessing.get_raster_info(
        lulc_snapshot_list[0])['raster_size']
    n_pixels_total = (n_rows * n_cols) * len(lulc_snapshot_list)
    n_pixels_processed = 0

    raster_tuple_list = []
    for raster_path in lulc_snapshot_list:
        raster = gdal.OpenEx(raster_path, gdal.OF_RASTER)
        band = raster.GetRasterBand(1)
        nodata = band.GetNoDataValue()
        raster_tuple_list.append((raster, band, nodata))

    transition_pairs = set()
    last_log_time = time.time()
    for block_offsets in pygeoprocessing.iterblocks((lulc_snapshot_list[0], 1),
                                                    offset_only=True):
        _, from_band, from_nodata = raster_tuple_list[0]
        from_array = from_band.ReadAsArray(**block_offsets)
        from_band = None

        for (_, to_band, to_nodata) in raster_tuple_list[1:]:
            if time.time() - last_log_time >= 5.0:
                percent_complete = n_pixels_processed / n_pixels_total
                LOGGER.info(
                    "Determining landcover transitions, "
                    f"{percent_complete:.2f}% complete.")

            to_array = to_band.ReadAsArray(**block_offsets)

            # This comparison assumes that our landcover rasters are of an
            # integer type.  When int matrices, we can compare directly to
            # None.
            valid_pixels = (
                ~utils.array_equals_nodata(from_array, from_nodata) &
                ~utils.array_equals_nodata(to_array, to_nodata))
            transition_pairs = transition_pairs.union(
                set(zip(from_array[valid_pixels].flatten(),
                        to_array[valid_pixels].flatten())))

            # Swap the arrays around to use the current 'to_array', 'to_nodata'
            # as the 'from_array', 'from_nodata' in the next iteration.
            from_array, from_nodata = (to_array, to_nodata)
            n_pixels_processed += to_array.size
    raster_tuple_list = None
    to_band = None
    LOGGER.info("Determining landcover transitions, 100.00%% complete.")

    # Mapping of whether the from, to landcover types are coastal blue carbon
    # habitats to the string carbon transition type.
    # The keys are structured as a tuple of two booleans where:
    #  * tuple[0] = whether the FROM transition is CBC habitat
    #  * tuple[1] = whether the TO transition is CBC habitat
    transition_types = {
        (True, True): 'accum',  # veg --> veg
        (False, True): 'accum',  # non-veg --> veg
        (True, False): 'disturb',  # veg --> non-veg
        (False, False): 'NCC',  # non-veg --> non-veg
    }

    sparse_transition_table = {}
    for from_lucode, to_lucode in transition_pairs:
        try:
            from_is_cbc = landcover_table[
                from_lucode]['is_coastal_blue_carbon_habitat']
            to_is_cbc = landcover_table[
                to_lucode]['is_coastal_blue_carbon_habitat']
        except KeyError:
            for variable in (from_lucode, to_lucode):
                if variable not in landcover_table:
                    raise ValueError(
                        'The landcover table is missing a row with the '
                        f'landuse code {variable}.')

        sparse_transition_table[(from_lucode, to_lucode)] = (
            transition_types[(from_is_cbc, to_is_cbc)])

    code_list = sorted([code for code in landcover_table.keys()])
    lulc_class_list_sorted = [
        landcover_table[code]['lulc-class'] for code in code_list]
    with open(target_table_path, 'w') as csv_file:
        fieldnames = ['lulc-class'] + lulc_class_list_sorted
        csv_file.write(f"{','.join(fieldnames)}\n")
        for row_code in code_list:
            class_name = landcover_table[row_code]['lulc-class']
            row = [class_name]
            for col_code in code_list:
                try:
                    column_value = sparse_transition_table[
                        (row_code, col_code)]
                except KeyError:
                    # When there isn't a transition that we know about, just
                    # leave the table blank.
                    column_value = ''
                row.append(column_value)
            csv_file.write(','.join(row) + '\n')

        # Append legend
        csv_file.write("\n,legend")
        csv_file.write(
            "\n,empty cells indicate that no transitions occur of that type")
        csv_file.write("\n,disturb (disturbance): change to low- med- or "
                       "high-impact-disturb")
        csv_file.write("\n,accum (accumulation)")
        csv_file.write("\n,NCC (no-carbon-change)")


def _create_biophysical_table(landcover_table, target_biophysical_table_path):
    """Write the biophysical table template to disk.

    The biophysical table templates contains all of the fields required by the
    main Coastal Blue Carbon model, and any field values that exist in the
    landcover table provided to this model will be carried over to the new
    table.

    Args:
        landcover_table (dict): A dict mapping int landcover codes to a dict
            with string keys that map to numeric or string column values.
        target_biophysical_table_path (string): The path to where the
            biophysical table template will be stored on disk.

    Returns:
        ``None``
    """
    target_column_names = [
        colname.lower() for colname in coastal_blue_carbon.ARGS_SPEC['args'][
            'biophysical_table_path']['columns']]

    with open(target_biophysical_table_path, 'w') as bio_table:
        bio_table.write(f"{','.join(target_column_names)}\n")
        for lulc_code in sorted(landcover_table.keys()):
            # 2 columns are defined below, and we need 1 less comma to only
            # have commas between fields.
            row = []
            for colname in target_column_names:
                try:
                    # Use the user's defined value if it exists
                    row.append(str(landcover_table[lulc_code][colname]))
                except KeyError:
                    row.append('')
            bio_table.write(f"{','.join(row)}\n")


@validation.invest_validator
def validate(args, limit_to=None):
    """Validate an input dictionary for Coastal Blue Carbon: Preprocessor.

    Args:
        args (dict): The args dictionary.
        limit_to=None (str or None): If a string key, only this args parameter
            will be validated.  If ``None``, all args parameters will be
            validated.

    Returns:
        A list of tuples where tuple[0] is an iterable of keys that the error
        message applies to and tuple[1] is the string validation warning.
    """
    return validation.validate(args, ARGS_SPEC['args'])



# coding=UTF-8
"""InVEST Habitat Quality model."""
import collections
import os
import logging

import numpy
from osgeo import gdal
from osgeo import osr
import pygeoprocessing
import taskgraph

from . import utils
from . import spec_utils
from .spec_utils import u
from . import validation
from . import MODEL_METADATA

LOGGER = logging.getLogger(__name__)

MISSING_SENSITIVITY_TABLE_THREATS_MSG = _(
    'Threats {threats} does not match any column in the sensitivity table. '
    'Sensitivity columns: {column_names}')  # (set of missing threats, set of found columns)
MISSING_COLUMN_MSG = _(
    "The column '{column_name}' was not found in the Threat Data table for "
    "the corresponding input LULC scenario.")
MISSING_THREAT_RASTER_MSG = _(
    "A threat raster for threats: {threat_list} was not found or it "
    "could not be opened by GDAL.")
DUPLICATE_PATHS_MSG = _("Threat paths must be unique. Duplicates: ")

ARGS_SPEC = {
    "model_name": MODEL_METADATA["habitat_quality"].model_title,
    "pyname": MODEL_METADATA["habitat_quality"].pyname,
    "userguide_html": MODEL_METADATA["habitat_quality"].userguide,
    "args_with_spatial_overlap": {
        "spatial_keys": [
            "lulc_cur_path", "lulc_fut_path", "lulc_bas_path",
            "access_vector_path"],
    },
    "args": {
        "workspace_dir": spec_utils.WORKSPACE,
        "results_suffix": spec_utils.SUFFIX,
        "n_workers": spec_utils.N_WORKERS,
        "lulc_cur_path": {
            **spec_utils.LULC,
            "projected": True,
            "about": _(
                "Map of LULC at present. All values in this raster must "
                "have corresponding entries in the Sensitivity table."),
            "name": _("current land cover")
        },
        "lulc_fut_path": {
            **spec_utils.LULC,
            "projected": True,
            "required": False,
            "about": _(
                "Map of LULC in a future scenario. All values in this raster "
                "must have corresponding entries in the Sensitivity "
                "Table. Must use the same classification scheme and codes as "
                "in the Current LULC map."),
            "name": _("future land cover")
        },
        "lulc_bas_path": {
            **spec_utils.LULC,
            "projected": True,
            "required": False,
            "about": _(
                "Map of LULC in a baseline scenario, when intensive landscape "
                "management was relatively rare. All values in this raster "
                "must have corresponding entries in the Sensitivity "
                "table. Must use the same classification scheme and codes as "
                "in the Current LULC map."),
            "name": _("baseline land cover")
        },
        "threats_table_path": {
            "type": "csv",
            "columns": {
                "threat": {
                    "type": "freestyle_string",
                    "about": _(
                        "Name of the threat. Each threat name must have a "
                        "corresponding column in the Sensitivity table.")},
                "max_dist": {
                    "type": "number",
                    "units": u.kilometer,
                    "about": _(
                        "The maximum distance over which each threat affects "
                        "habitat quality. The impact of each degradation "
                        "source will decline to zero at this maximum "
                        "distance. This value must be greater than or equal "
                        "to the pixel size of your LULC raster(s).")
                },
                "weight": {
                    "type": "ratio",
                    "about": _(
                        "The impact of each threat on habitat quality, "
                        "relative to other threats.")
                },
                "decay": {
                    "type": "option_string",
                    "options": {
                        "linear": {
                            "description": _(
                                "Effects of the threat decay linearly with "
                                "distance from the threat.")},
                        "exponential": {
                            "description": _(
                                "Effects of the threat decay exponentially "
                                "with distance from the threat.")}
                    },
                    "about": _("The type of decay over space for each threat.")
                },
                "cur_path": {
                    "type": "raster",
                    "bands": {1: {"type": "ratio"}},
                    "about": _(
                        "Map of the threat's distribution in the current "
                        "scenario. Each pixel value is the relative intensity "
                        "of the threat at that location. ")
                },
                "fut_path": {
                    "required": "lulc_fut_path",
                    "type": "raster",
                    "bands": {1: {"type": "ratio"}},
                    "about": _(
                        "Map of the threat's distribution in the future "
                        "scenario. Each pixel value is the relative intensity "
                        "of the threat at that location. "
                        "Required if Future LULC is provided.")
                },
                "base_path": {
                    "required": "lulc_bas_path",
                    "type": "raster",
                    "bands": {1: {"type": "ratio"}},
                    "about": _(
                        "Map of the threat's distribution in the baseline "
                        "scenario. Each pixel value is the relative intensity "
                        "of the threat at that location. "
                        "Required if Baseline LULC is provided.")
                }
            },
            "about": _(
                "Table mapping each threat of interest to its properties and "
                "distribution maps. Paths are relative to the threats "
                "table path."),
            "name": _("threats table")
        },
        "access_vector_path": {
            "type": "vector",
            "projected": True,
            "fields": {
                "access": {
                    "type": "ratio",
                    "about": _(
                        "The region's relative accessibility to threats, "
                        "where 0 represents completely inaccessible and 1 "
                        "represents completely accessible.")
                }
            },
            "geometries": spec_utils.POLYGONS,
            "required": False,
            "about": _(
                "Map of the relative protection that legal, institutional, "
                "social, and physical barriers provide against threats. Any "
                "cells not covered by a polygon will be set to 1."),
            "name": _("accessibility to threats")
        },
        "sensitivity_table_path": {
            "type": "csv",
            "columns": {
                "lulc": {
                    "type": "integer",
                    "about": _("LULC codes corresponding to those in the LULC "
                               "rasters.")
                },
                "habitat": {
                    "type": "ratio",
                    "about": _(
                        "Suitability of this LULC class as habitat, where 0 "
                        "is not suitable and 1 is completely suitable.")
                },
                "[THREAT]": {
                    "type": "ratio",
                    "about": _(
                        "The relative sensitivity of each LULC class to each "
                        "type of threat, where 1 represents high sensitivity "
                        "and 0 represents that it is unaffected. There must "
                        "be one threat column for each threat name in the "
                        "'threats' column of the Threats Table.")
                }
            },
            "about": _(
                "Table mapping each LULC class to data about the species' "
                "habitat preference and threat sensitivity in areas with that "
                "LULC."),
            "name": _("sensitivity table")
        },
        "half_saturation_constant": {
            "expression": "value > 0",
            "type": "number",
            "units": u.none,
            "about": _(
                "Half-saturation constant used in the degradation equation."),
            "name": _("half-saturation constant")
        },
    }
}
# All out rasters besides rarity should be gte to 0. Set nodata accordingly.
_OUT_NODATA = float(numpy.finfo(numpy.float32).min)
# Scaling parameter from User's Guide eq. 4 for quality of habitat
_SCALING_PARAM = 2.5
# To help track and name threat rasters from paths in threat table columns
_THREAT_SCENARIO_MAP = {'_c': 'cur_path', '_f': 'fut_path', '_b': 'base_path'}


def execute(args):
    """Habitat Quality.

    This model calculates habitat degradation and quality for the current LULC
    as described in the InVEST user's guide. Optionally ``execute`` calculates
    habitat degradation and quality for a future LULC and habitat rarity for
    current and future LULC.

    Args:
        args (dict): a key, value mapping for the habitat quality inputs.
        args['workspace_dir'] (string): a path to the directory that will
            write output and other temporary files (required)
        args['lulc_cur_path'] (string): a path to an input land use/land
            cover raster (required)
        args['lulc_fut_path'] (string): a path to an input land use/land
            cover raster (optional)
        args['lulc_bas_path'] (string): a path to an input land use/land
            cover raster (optional, but required for rarity calculations)
        args['threats_table_path'] (string): a path to an input CSV
            containing data of all the considered threats. Each row is a
            degradation source and each column a different attribute of the
            source with the following names (case-insensitive):
            'THREAT','MAX_DIST','WEIGHT', 'DECAY', 'CUR_PATH' (required)
            'BASE_PATH', 'FUT_PATH' (optional).
        args['access_vector_path'] (string): a path to an input polygon
            shapefile containing data on the relative protection against
            threats (optional)
        args['sensitivity_table_path'] (string): a path to an input CSV file
            of LULC types, whether they are considered habitat, and their
            sensitivity to each threat (required)
        args['half_saturation_constant'] (float): a python float that
            determines the spread and central tendency of habitat quality
            scores (required)
        args['results_suffix'] (string): a python string that will be inserted
            into all raster path paths just before the file extension.
        args['n_workers'] (int): (optional) The number of worker processes to
            use for processing this model.  If omitted, computation will take
            place in the current process.

    Returns:
        None
    """
    LOGGER.info("Starting execute of Habitat Quality model.")
    # Append a _ to the suffix if it's not empty and doesn't already have one
    file_suffix = utils.make_suffix_string(args, 'results_suffix')

    # Check to see if each of the workspace folders exists. If not, create the
    # folder in the filesystem.
    LOGGER.info("Creating workspace")
    output_dir = args['workspace_dir']
    intermediate_output_dir = os.path.join(
        args['workspace_dir'], 'intermediate')
    kernel_dir = os.path.join(intermediate_output_dir, 'kernels')
    utils.make_directories([intermediate_output_dir, output_dir, kernel_dir])

    taskgraph_working_dir = os.path.join(
        intermediate_output_dir, '_taskgraph_working_dir')

    n_workers = int(args.get('n_workers', -1))
    task_graph = taskgraph.TaskGraph(taskgraph_working_dir, n_workers)

    LOGGER.info("Checking Threat and Sensitivity tables for compliance")
    # Get CSVs as dictionaries and ensure the key is a string for threats.
    threat_dict = {
        str(key): value for key, value in utils.build_lookup_from_csv(
            args['threats_table_path'], 'THREAT', to_lower=True).items()}
    sensitivity_dict = utils.build_lookup_from_csv(
        args['sensitivity_table_path'], 'LULC', to_lower=True)

    # Get the directory path for the Threats CSV, used for locating threat
    # rasters, which are relative to this path
    threat_csv_dirpath = os.path.dirname(args['threats_table_path'])

    half_saturation_constant = float(args['half_saturation_constant'])

    # Dictionary for reclassing habitat values
    sensitivity_reclassify_habitat_dict = {
        int(key): float(val['habitat']) for key, val in
        sensitivity_dict.items()}

    # declare dictionaries to store the land cover and the threat rasters
    # pertaining to the different threats
    lulc_path_dict = {}
    threat_path_dict = {}
    # store land cover and threat rasters in a list for convenient access
    lulc_and_threat_raster_list = []
    # list for checking threat values tasks
    threat_values_task_lookup = {}
    LOGGER.info("Validate threat rasters and collect unique LULC codes")
    # compile all the threat rasters associated with the land cover
    for lulc_key, lulc_arg in (('_c', 'lulc_cur_path'),
                               ('_f', 'lulc_fut_path'),
                               ('_b', 'lulc_bas_path')):
        if lulc_arg in args and args[lulc_arg] != '':
            lulc_path = args[lulc_arg]
            lulc_path_dict[lulc_key] = lulc_path
            # save land cover paths in a list for alignment and resize
            lulc_and_threat_raster_list.append(lulc_path)

            # add a key to the threat dictionary that associates all threat
            # rasters with this land cover
            threat_path_dict['threat' + lulc_key] = {}

            # for each threat given in the CSV file try opening the associated
            # raster which should be found relative to the Threat CSV
            for threat in threat_dict:
                LOGGER.debug(f"Validating path for threat: {threat}")
                # Build absolute threat path from threat table
                threat_table_path_col = _THREAT_SCENARIO_MAP[lulc_key]
                threat_path_relative = (
                    threat_dict[threat][threat_table_path_col])
                threat_path = os.path.join(
                    threat_csv_dirpath, threat_path_relative)

                threat_path_err_msg = (
                    'There was an Error locating a threat raster from '
                    'the path in CSV for column: '
                    f'{_THREAT_SCENARIO_MAP[lulc_key]} and threat: '
                    f'{threat}. The path in the CSV column should be '
                    'relative to the threat CSV table.')

                threat_validate_result = _validate_threat_path(
                    threat_path, lulc_key)
                if threat_validate_result == 'error':
                    raise ValueError(threat_path_err_msg)

                threat_path = threat_validate_result

                threat_path_dict['threat' + lulc_key][threat] = threat_path
                # save threat paths in a list for alignment and resize
                if threat_path:
                    # check for duplicate absolute threat path names that
                    # cause errors when trying to write aligned versions
                    if (threat_path not in lulc_and_threat_raster_list):
                        lulc_and_threat_raster_list.append(threat_path)
                    else:
                        raise ValueError(
                            DUPLICATE_PATHS_MSG + os.path.basename(threat_path)
                        )
                    # Check threat raster values are 0 <= x <= 1
                    threat_values_task = task_graph.add_task(
                        func=_raster_values_in_bounds,
                        args=((threat_path, 1), 0.0, 1.0),
                        store_result=True,
                        task_name=f'check_threat_values{lulc_key}_{threat}')
                    threat_values_task_lookup[threat_values_task.task_name] = {
                        'task': threat_values_task,
                        'path': threat_path_relative,
                        'table_col': threat_table_path_col}

    LOGGER.info("Checking threat raster values are valid ( 0 <= x <= 1 ).")
    # Assert that threat rasters have valid values.
    for values in threat_values_task_lookup.values():
        # get returned boolean to see if values were valid
        valid_threat_values = values['task'].get()
        if not valid_threat_values:
            raise ValueError(
                "Threat rasters should have values between 0 and 1, however,"
                f"Threat: {values['path']} for column: {values['table_col']}",
                " had values outside of this range.")

    LOGGER.info('Aligning and resizing land cover and threat rasters')
    lulc_raster_info = pygeoprocessing.get_raster_info(args['lulc_cur_path'])
    # ensure that the pixel size used is square
    lulc_pixel_size = lulc_raster_info['pixel_size']
    min_pixel_size = min([abs(x) for x in lulc_pixel_size])
    pixel_size = (min_pixel_size, -min_pixel_size)
    lulc_bbox = lulc_raster_info['bounding_box']

    # create paths for aligned rasters checking for the case the raster path
    # is a folder
    aligned_raster_list = []
    for path in lulc_and_threat_raster_list:
        if os.path.isdir(path):
            threat_dir_name = (f'{os.path.basename(os.path.dirname(path))}'
                               f'_aligned{file_suffix}.tif')
            aligned_raster_list.append(
                os.path.join(intermediate_output_dir, threat_dir_name))
        else:
            aligned_path = (f'{os.path.splitext(os.path.basename(path))[0]}'
                            f'_aligned{file_suffix}.tif')
            aligned_raster_list.append(
                os.path.join(intermediate_output_dir, aligned_path))

    LOGGER.debug(f"Raster paths for aligning: {aligned_raster_list}")
    # Align and resize all the land cover and threat rasters,
    # and store them in the intermediate folder
    align_task = task_graph.add_task(
        func=pygeoprocessing.align_and_resize_raster_stack,
        args=(
            lulc_and_threat_raster_list, aligned_raster_list,
            ['near']*len(lulc_and_threat_raster_list), pixel_size,
            lulc_bbox),
        target_path_list=aligned_raster_list,
        task_name='align_input_rasters')

    LOGGER.debug("Updating dict raster paths to reflect aligned paths")
    # Modify paths in lulc_path_dict and threat_path_dict to be aligned rasters
    for lulc_key, lulc_path in lulc_path_dict.items():
        lulc_path_dict[lulc_key] = os.path.join(
            intermediate_output_dir,
            (f'{os.path.splitext(os.path.basename(lulc_path))[0]}'
             f'_aligned{file_suffix}.tif'))
        for threat in threat_dict:
            threat_path = threat_path_dict['threat' + lulc_key][threat]
            if threat_path in lulc_and_threat_raster_list:
                aligned_threat_path = os.path.join(
                    intermediate_output_dir,
                    (f'{os.path.splitext(os.path.basename(threat_path))[0]}'
                     f'_aligned{file_suffix}.tif'))
                # Use these updated threat raster paths in future calculations
                threat_path_dict['threat' + lulc_key][threat] = (
                    aligned_threat_path)

    LOGGER.info('Starting habitat_quality biophysical calculations')
    # Rasterize access vector, if value is null set to 1 (fully accessible),
    # else set to the value according to the ACCESS attribute
    cur_lulc_path = lulc_path_dict['_c']

    access_raster_path = os.path.join(
        intermediate_output_dir, f'access_layer{file_suffix}.tif')
    # create a new raster based on the raster info of current land cover.
    # fill with 1.0 for case where no access shapefile provided,
    # which indicates we don't want to mask anything out later
    create_access_raster_task = task_graph.add_task(
        func=pygeoprocessing.new_raster_from_base,
        args=(cur_lulc_path, access_raster_path, gdal.GDT_Float32,
              [_OUT_NODATA]),
        kwargs={
            'fill_value_list': [1.0]
        },
        target_path_list=[access_raster_path],
        dependent_task_list=[align_task],
        task_name='access_raster')
    access_task_list = [create_access_raster_task]

    if 'access_vector_path' in args and args['access_vector_path']:
        LOGGER.debug("Rasterize Access vector")
        rasterize_access_task = task_graph.add_task(
            func=pygeoprocessing.rasterize,
            args=(args['access_vector_path'], access_raster_path),
            kwargs={
                'option_list': ['ATTRIBUTE=ACCESS'],
                'burn_values': None
            },
            target_path_list=[access_raster_path],
            dependent_task_list=[create_access_raster_task],
            task_name='rasterize_access')
        access_task_list.append(rasterize_access_task)

    # calculate the weight sum which is the sum of all the threats' weights
    weight_sum = 0.0
    for threat_data in threat_dict.values():
        # Sum weight of threats
        weight_sum = weight_sum + threat_data['weight']

    # for each land cover raster provided compute habitat quality
    for lulc_key, lulc_path in lulc_path_dict.items():
        LOGGER.info(f'Calculating habitat quality for landuse: {lulc_path}')

        threat_convolve_task_list = []
        sensitivity_task_list = []

        # Create raster of habitat based on habitat field
        habitat_raster_path = os.path.join(
            intermediate_output_dir,
            f'habitat{lulc_key}{file_suffix}.tif')

        reclass_error_details = {
            'raster_name': f'LULC{lulc_key}', 'column_name': 'lucode',
            'table_name': 'Sensitivity'}
        habitat_raster_task = task_graph.add_task(
            func=utils.reclassify_raster,
            args=((lulc_path, 1), sensitivity_reclassify_habitat_dict,
                  habitat_raster_path, gdal.GDT_Float32, _OUT_NODATA,
                  reclass_error_details),
            dependent_task_list=[align_task],
            task_name=f'habitat_raster{lulc_key}')

        # initialize a list that will store all the threat/threat rasters
        # after they have been adjusted for distance, weight, and access
        deg_raster_list = []

        # a list to keep track of the normalized weight for each threat
        weight_list = numpy.array([])

        # variable to indicate whether we should break out of calculations
        # for a land cover because a threat raster was not found
        exit_landcover = False

        # adjust each threat/threat raster for distance, weight, and access
        for threat, threat_data in threat_dict.items():
            LOGGER.debug(
                f'Calculating threat: {threat}.\nThreat data: {threat_data}')

            # get the threat raster for the specific threat
            threat_raster_path = threat_path_dict['threat' + lulc_key][threat]
            # if threat path is None then must be in Base scenario where
            # threats are not required.
            if threat_raster_path is None:
                LOGGER.warning(
                    f'The threat raster for {threat} could not be found for'
                    f' the land cover {lulc_key}. Skipping Habitat Quality'
                    ' calculation for this land cover.')
                exit_landcover = True
                break
            # Check to make sure max_dist is greater than 0
            if threat_data['max_dist'] <= 0.0:
                raise ValueError(
                    f"The max distance for threat: '{threat}' is less than"
                    " or equal to 0. MAX_DIST should be a positive value.")

            kernel_path = os.path.join(
                kernel_dir, f'kernel_{threat}{lulc_key}{file_suffix}.tif')

            decay_type = threat_data['decay']

            create_kernel_task = task_graph.add_task(
                func=_create_decay_kernel,
                args=((threat_raster_path, 1), kernel_path, decay_type,
                      threat_data['max_dist']),
                target_path_list=[kernel_path],
                dependent_task_list=[align_task],
                task_name=f'decay_kernel_{decay_type}{lulc_key}_{threat}')

            filtered_threat_raster_path = os.path.join(
                intermediate_output_dir,
                f'filtered_{threat}{lulc_key}{file_suffix}.tif')

            convolve_task = task_graph.add_task(
                func=pygeoprocessing.convolve_2d,
                args=((threat_raster_path, 1), (kernel_path, 1),
                      filtered_threat_raster_path),
                kwargs={
                    'target_nodata': _OUT_NODATA,
                    'ignore_nodata_and_edges': False,
                    'mask_nodata': False
                },
                target_path_list=[filtered_threat_raster_path],
                dependent_task_list=[create_kernel_task],
                task_name=f'convolve_{decay_type}{lulc_key}_{threat}')
            threat_convolve_task_list.append(convolve_task)

            # create sensitivity raster based on threat
            sens_raster_path = os.path.join(
                intermediate_output_dir,
                f'sens_{threat}{lulc_key}{file_suffix}.tif')

            # Dictionary for reclassing threat sensitivity values
            sensitivity_reclassify_threat_dict = {
                int(key): float(val[threat]) for key, val in
                sensitivity_dict.items()}

            reclass_error_details = {
                'raster_name': 'LULC', 'column_name': 'lucode',
                'table_name': 'Sensitivity'}
            sens_threat_task = task_graph.add_task(
                func=utils.reclassify_raster,
                args=((lulc_path, 1), sensitivity_reclassify_threat_dict,
                      sens_raster_path, gdal.GDT_Float32, _OUT_NODATA,
                      reclass_error_details),
                target_path_list=[sens_raster_path],
                dependent_task_list=[align_task],
                task_name=f'sens_raster_{decay_type}{lulc_key}_{threat}')
            sensitivity_task_list.append(sens_threat_task)

            # get the normalized weight for each threat
            weight_avg = threat_data['weight'] / weight_sum

            # add the threat raster adjusted by distance and the raster
            # representing sensitivity to the list to be past to
            # vectorized_rasters below
            deg_raster_list.append(filtered_threat_raster_path)
            deg_raster_list.append(sens_raster_path)

            # store the normalized weight for each threat in a list that
            # will be used below in total_degradation
            weight_list = numpy.append(weight_list, weight_avg)

        # check to see if we got here because a threat raster was missing
        # for baseline lulc, if so then we want to skip to the next landcover
        if exit_landcover:
            continue

        # add the access_raster onto the end of the collected raster list. The
        # access_raster will be values from the shapefile if provided or a
        # raster filled with all 1's if not
        deg_raster_list.append(access_raster_path)

        deg_sum_raster_path = os.path.join(
            output_dir, f'deg_sum{lulc_key}{file_suffix}.tif')

        LOGGER.info('Starting raster calculation on total degradation')

        total_degradation_task = task_graph.add_task(
            func=_calculate_total_degradation,
            args=(deg_raster_list, deg_sum_raster_path, weight_list),
            target_path_list=[deg_sum_raster_path],
            dependent_task_list=[
                *threat_convolve_task_list, *sensitivity_task_list,
                *access_task_list],
            task_name=f'tot_degradation_{decay_type}{lulc_key}_{threat}')

        # Compute habitat quality
        # ksq: a term used below to compute habitat quality
        ksq = half_saturation_constant**_SCALING_PARAM

        quality_path = os.path.join(
            output_dir, f'quality{lulc_key}{file_suffix}.tif')

        LOGGER.info('Starting raster calculation on quality')

        deg_hab_raster_list = [deg_sum_raster_path, habitat_raster_path]

        _ = task_graph.add_task(
            func=_calculate_habitat_quality,
            args=(deg_hab_raster_list, quality_path, ksq),
            target_path_list=[quality_path],
            dependent_task_list=[habitat_raster_task, total_degradation_task],
            task_name='habitat_quality')

    # Compute Rarity if user supplied baseline raster
    if '_b' not in lulc_path_dict:
        LOGGER.info('Baseline not provided to compute Rarity')
    else:
        lulc_base_path = lulc_path_dict['_b']

        # compute rarity for current landscape and future (if provided)
        for lulc_key in ['_c', '_f']:
            if lulc_key not in lulc_path_dict:
                continue
            lulc_path = lulc_path_dict[lulc_key]
            lulc_time = 'current' if lulc_key == '_c' else 'future'

            new_cover_path = os.path.join(
                intermediate_output_dir,
                f'new_cover{lulc_key}{file_suffix}.tif')

            rarity_path = os.path.join(
                output_dir, f'rarity{lulc_key}{file_suffix}.tif')

            _ = task_graph.add_task(
                func=_compute_rarity_operation,
                args=((lulc_base_path, 1), (lulc_path, 1), (new_cover_path, 1),
                      rarity_path),
                dependent_task_list=[align_task],
                task_name=f'rarity{lulc_time}')

    task_graph.close()
    task_graph.join()
    LOGGER.info("Habitat Quality Model complete.")


def _calculate_habitat_quality(deg_hab_raster_list, quality_out_path, ksq):
    """Calculate habitat quality from degradation inputs.

    Args:
        deg_hab_raster_list (list): list of string paths for the degraded
            habitat rasters.
        quality_out_path (string): path to output the habitat quality raster.
        ksq (float): a number representing half-saturation**_SCALING_PARAM

    Returns:
        None
    """
    def quality_op(degradation, habitat):
        """Computes habitat quality given degradation and habitat values."""
        out_array = numpy.empty_like(degradation)
        out_array[:] = _OUT_NODATA
        # Both these rasters are Float32, so the actual pixel values written
        # might be *slightly* off of _OUT_NODATA but should still be
        # interpreted as nodata.
        # _OUT_NODATA (defined above) should never be None, so this is okay
        valid_pixels = ~(
            utils.array_equals_nodata(degradation, _OUT_NODATA) |
            utils.array_equals_nodata(habitat, _OUT_NODATA))

        out_array[valid_pixels] = (
            habitat[valid_pixels] *
            (1.0 - (degradation[valid_pixels]**_SCALING_PARAM) /
                (degradation[valid_pixels]**_SCALING_PARAM + ksq)))
        return out_array

    deg_hab_raster_band_list = [
        (path, 1) for path in deg_hab_raster_list]

    pygeoprocessing.raster_calculator(
        deg_hab_raster_band_list, quality_op, quality_out_path,
        gdal.GDT_Float32, _OUT_NODATA)


def _calculate_total_degradation(
        deg_raster_list, deg_sum_raster_path, weight_list):
    """Calculate habitat degradation.

    Args:
        deg_raster_list (list): list of string paths for the degraded
            threat rasters.
        deg_sum_raster_path (string): path to output the habitat quality
            degradation raster.
        weight_list (list): normalized weight for each threat corresponding
            to threats in ``deg_raster_list``.

    Returns:
        None
    """
    def total_degradation(*raster):
        """Computes the total degradation value.

        Args:
            *raster (list): a list of numpy arrays of float type depicting
                the adjusted threat value per pixel based on distance and
                sensitivity. The values are in pairs so that the values for
                each threat can be tracked:
                [filtered_val_threat1, sens_val_threat1,
                 filtered_val_threat2, sens_val_threat2, ...]
                There is an optional last value in the list which is the
                access_raster value, but it is only present if
                access_raster is not None.

        Returns:
            The total degradation score for the pixel.
        """
        # we can not be certain how many threats the user will enter,
        # so we handle each filtered threat and sensitivity raster
        # in pairs
        sum_degradation = numpy.zeros(raster[0].shape)
        for index in range(len(raster) // 2):
            step = index * 2
            sum_degradation += (
                raster[step] * raster[step + 1] * weight_list[index])

        nodata_mask = numpy.empty(raster[0].shape, dtype=numpy.int8)
        nodata_mask[:] = 0
        for array in raster:
            nodata_mask = nodata_mask | utils.array_equals_nodata(
                array, _OUT_NODATA)

        # the last element in raster is access
        return numpy.where(
            nodata_mask, _OUT_NODATA, sum_degradation * raster[-1])

    deg_raster_band_list = [(path, 1) for path in deg_raster_list]

    pygeoprocessing.raster_calculator(
        deg_raster_band_list, total_degradation, deg_sum_raster_path,
        gdal.GDT_Float32, _OUT_NODATA)


def _compute_rarity_operation(
        base_lulc_path_band, lulc_path_band, new_cover_path, rarity_path):
    """Calculate habitat rarity.

    Output rarity values will be an index from 0 - 1 where:
       pixel > 0.5 - more rare
       pixel < 0.5 - less rare
       pixel = 0.5 - no rarity change
       pixel = 0.0 - LULC not found in the baseline for comparison

    Args:
        base_lulc_path_band (tuple): a 2 tuple for the path to input base
            LULC raster of the form (path, band index).
        lulc_path_band (tuple):  a 2 tuple for the path to LULC for current
            or future scenario of the form (path, band index).
        new_cover_path (tuple): a 2 tuple for the path to intermediate
            raster file for trimming ``lulc_path_band`` to
            ``base_lulc_path_band`` of the form (path, band index).
        rarity_path (string): path to output rarity raster.

    Returns:
        None
    """
    # get the area of a base pixel to use for computing rarity where the
    # pixel sizes are different between base and cur/fut rasters
    base_raster_info = pygeoprocessing.get_raster_info(
        base_lulc_path_band[0])
    base_pixel_size = base_raster_info['pixel_size']
    base_area = float(abs(base_pixel_size[0]) * abs(base_pixel_size[1]))
    base_nodata = base_raster_info['nodata'][0]

    lulc_code_count_b = _raster_pixel_count(base_lulc_path_band)

    # get the area of a cur/fut pixel
    lulc_raster_info = pygeoprocessing.get_raster_info(lulc_path_band[0])
    lulc_pixel_size = lulc_raster_info['pixel_size']
    lulc_area = float(abs(lulc_pixel_size[0]) * abs(lulc_pixel_size[1]))
    lulc_nodata = lulc_raster_info['nodata'][0]

    def trim_op(base, cover_x):
        """Trim cover_x to the mask of base.

        Args:
            base (numpy.ndarray): base raster from 'lulc_base'
            cover_x (numpy.ndarray): either future or current land
                cover raster from 'lulc_path_band' above

        Returns:
            _OUT_NODATA where either array has nodata, otherwise cover_x.
        """
        result_array = numpy.full(cover_x.shape, _OUT_NODATA)
        valid_mask = (
            ~utils.array_equals_nodata(base, base_nodata) &
            ~utils.array_equals_nodata(cover_x, lulc_nodata))
        result_array[valid_mask] = cover_x[valid_mask]
        return result_array

    pygeoprocessing.raster_calculator(
        [base_lulc_path_band, lulc_path_band], trim_op, new_cover_path[0],
        gdal.GDT_Float32, _OUT_NODATA)

    LOGGER.info('Starting rarity computation on'
                f' {os.path.basename(lulc_path_band[0])} land cover.')

    lulc_code_count_x = _raster_pixel_count(new_cover_path)

    # a dictionary to map LULC types to a number that depicts how
    # rare they are considered
    code_index = {}

    # compute rarity index for each lulc code
    # define 0.0 if an lulc code is found in the cur/fut landcover
    # but not the baseline
    for code in lulc_code_count_x:
        if code in lulc_code_count_b:
            numerator = lulc_code_count_x[code] * lulc_area
            denominator = lulc_code_count_b[code] * base_area
            ratio = 1.0 - (numerator / (denominator + numerator))
            code_index[code] = ratio
        else:
            code_index[code] = 0.0

    pygeoprocessing.reclassify_raster(
        new_cover_path, code_index, rarity_path, gdal.GDT_Float32,
        _OUT_NODATA)

    LOGGER.info('Finished rarity computation on'
                f' {os.path.basename(lulc_path_band[0])} land cover.')


def _create_decay_kernel(raster_path_band, kernel_path, decay_type, max_dist):
    """Create a decay kernel as a raster.

    Args:
        raster_path_band (tuple): a 2 tuple of the form
            (filepath to raster, band index) for raster to decay.
        kernel_path (string): path to output kernel raster.
        decay_type (string): type of decay kernel to create, either
            'linear' | 'exponentional'.
        max_dist (float): max distance of threat in KM.

    Returns:
        None
    """
    # need the pixel size for the raster so we can create an appropriate
    # kernel for convolution
    threat_pixel_size = pygeoprocessing.get_raster_info(
        raster_path_band[0])['pixel_size']

    # convert max distance (given in KM) to meters
    max_dist_m = max_dist * 1000.0

    # convert max distance from meters to the number of pixels that
    # represents on the raster
    max_dist_pixel = max_dist_m / abs(threat_pixel_size[0])
    LOGGER.debug(f'Max distance in pixels: {max_dist_pixel}')

    # blur the raster based on the decay type
    if decay_type == 'linear':
        decay_func = _make_linear_decay_kernel_path
    elif decay_type == 'exponential':
        decay_func = utils.exponential_decay_kernel_raster
    else:
        raise ValueError(
            "Unknown type of decay in biophysical table, should be"
            f" either 'linear' or 'exponential'. Input was {decay_type}"
            " for threat"
            f" {os.path.splitext(os.path.basename(raster_path_band[0]))[0]}")

    decay_func(max_dist_pixel, kernel_path)


def _raster_pixel_count(raster_path_band):
    """Count unique pixel values in single band raster.

    Args:
        raster_path_band (tuple): a 2 tuple of the form
            (filepath to raster, band index) where the raster has a single
            band.

    Returns:
        dict of pixel values to frequency.
    """
    nodata = pygeoprocessing.get_raster_info(
        raster_path_band[0])['nodata'][raster_path_band[1]-1]

    counts = collections.defaultdict(int)
    for _, raster_block in pygeoprocessing.iterblocks(raster_path_band):
        for value, count in zip(
                *numpy.unique(raster_block, return_counts=True)):
            if value == nodata:
                continue
            counts[value] += count
    return counts


def _make_linear_decay_kernel_path(max_distance, kernel_path):
    """Create a linear decay kernel as a raster.

    Pixels in raster are equal to d / max_distance where d is the distance to
    the center of the raster in number of pixels.

    Args:
        max_distance (int): number of pixels out until the decay is 0.
        kernel_path (string): path to output raster whose values are in (0,1)
            representing distance to edge.
            Size is (``max_distance`` * 2 + 1)

    Returns:
        None
    """
    kernel_size = int(numpy.round(max_distance * 2 + 1))

    driver = gdal.GetDriverByName('GTiff')
    kernel_dataset = driver.Create(
        kernel_path.encode('utf-8'), kernel_size, kernel_size, 1,
        gdal.GDT_Float32, options=['BIGTIFF=IF_SAFER'])

    # Make some kind of geotransform, it doesn't matter what but
    # will make GIS libraries behave better if it's all defined
    kernel_dataset.SetGeoTransform([444720, 30, 0, 3751320, 0, -30])
    srs = osr.SpatialReference()
    srs.SetUTM(11, 1)
    srs.SetWellKnownGeogCS('NAD27')
    kernel_dataset.SetProjection(srs.ExportToWkt())

    kernel_band = kernel_dataset.GetRasterBand(1)
    kernel_band.SetNoDataValue(-9999)

    col_index = numpy.array(range(kernel_size))
    integration = 0.0
    for row_index in range(kernel_size):
        distance_kernel_row = numpy.sqrt(
            (row_index - max_distance) ** 2 +
            (col_index - max_distance) ** 2).reshape(1, kernel_size)
        kernel = numpy.where(
            distance_kernel_row > max_distance, 0.0,
            (max_distance - distance_kernel_row) / max_distance)
        integration += numpy.sum(kernel)
        kernel_band.WriteArray(kernel, xoff=0, yoff=row_index)

    for row_index in range(kernel_size):
        kernel_row = kernel_band.ReadAsArray(
            xoff=0, yoff=row_index, win_xsize=kernel_size, win_ysize=1)
        kernel_row /= integration
        kernel_band.WriteArray(kernel_row, 0, row_index)


def _raster_values_in_bounds(raster_path_band, lower_bound, upper_bound):
    """Check raster values are between ``lower_bound`` and ``upper_bound``.

    Check that the raster has values ``lower_bound`` <= x <= ``upper_bound``.
    Nodata values are ignored.

    Args:
        raster_path_band (tuple): a 2 tuple for a GDAL raster path with
            the form (filepath, band index) to the raster on disk.
        lower_bound (int): integer for the lower bound of raster values,
            inclusive.
        upper_bound (int): integer for the upper bound of raster values,
            inclusive.

    Returns:
        True if values are within range and False otherwise.
    """
    raster_info = pygeoprocessing.get_raster_info(raster_path_band[0])
    raster_nodata = raster_info['nodata'][0]

    if raster_nodata is None:
        LOGGER.warning(
            f"Raster has undefined NODATA value for {raster_path_band[0]}.")
        # If raster nodata is None then set to _OUT_NODATA to use for masking
        # where in this case nodata_mask will be all False.
        raster_nodata = _OUT_NODATA

    values_valid = True

    for _, raster_block in pygeoprocessing.iterblocks(raster_path_band):
        nodata_mask = ~utils.array_equals_nodata(raster_block, raster_nodata)
        if ((raster_block[nodata_mask] < lower_bound) |
                (raster_block[nodata_mask] > upper_bound)).any():
            values_valid = False
            break

    return values_valid


def _validate_threat_path(threat_path, lulc_key):
    """Check ``threat_path`` is a valid raster file against ``lulc_key``.

    Check to see that the path is a valid raster and if not use ``lulc_key``
    to determine how to handle the non valid raster.

    Args:
        threat_path (str): path on disk for a possible raster file.
        lulc_key (str): an string indicating which land cover this threat
            path is associated with. Can be: '_b' | '_c' | '_f'

    Returns:
        If ``threat_path`` is a valid raster file then,
            return ``threat_path``.
        If ``threat_path`` is not valid then,
            return ``None`` if ``lulc_key`` == '_b'
            return 'error` otherwise
    """
    # Checking threat path exists to control custom error messages
    # for user readability.
    if os.path.exists(threat_path):
        threat_gis_type = pygeoprocessing.get_gis_type(threat_path)
        if threat_gis_type != pygeoprocessing.RASTER_TYPE:
            # Raise a value error with custom message to help users
            # debug threat raster issues
            if lulc_key != '_b':
                return "error"
            # it's OK to have no threat raster w/ baseline scenario
            else:
                return None
        else:
            return threat_path
    else:
        if lulc_key != '_b':
            return "error"
        else:
            return None


@validation.invest_validator
def validate(args, limit_to=None):
    """Validate args to ensure they conform to ``execute``'s contract.

    Args:
        args (dict): dictionary of key(str)/value pairs where keys and
            values are specified in ``execute`` docstring.
        limit_to (str): (optional) if not None indicates that validation
            should only occur on the args[limit_to] value. The intent that
            individual key validation could be significantly less expensive
            than validating the entire ``args`` dictionary.

    Returns:
        list of ([invalid key_a, invalid_keyb, ...], 'warning/error message')
            tuples. Where an entry indicates that the invalid keys caused
            the error message in the second part of the tuple. This should
            be an empty list if validation succeeds.
    """
    validation_warnings = validation.validate(
        args, ARGS_SPEC['args'], ARGS_SPEC['args_with_spatial_overlap'])

    invalid_keys = validation.get_invalid_keys(validation_warnings)

    if ("threats_table_path" not in invalid_keys and
            "sensitivity_table_path" not in invalid_keys and
            "threat_raster_folder" not in invalid_keys):

        # Get CSVs as dictionaries and ensure the key is a string for threats.
        threat_dict = {
            str(key): value for key, value in utils.build_lookup_from_csv(
                args['threats_table_path'], 'THREAT', to_lower=True).items()}
        sensitivity_dict = utils.build_lookup_from_csv(
            args['sensitivity_table_path'], 'LULC', to_lower=True)

        # check that the threat names in the threats table match with the
        # threats columns in the sensitivity table.
        sens_header_set = set(list(sensitivity_dict.values())[0])
        threat_set = {threat for threat in threat_dict}
        missing_sens_header_set = threat_set.difference(sens_header_set)

        if missing_sens_header_set:
            validation_warnings.append(
                (['sensitivity_table_path'],
                 MISSING_SENSITIVITY_TABLE_THREATS_MSG.format(
                    threats=missing_sens_header_set,
                    column_names=sens_header_set)))

            invalid_keys.add('sensitivity_table_path')

        # Get the directory path for the Threats CSV, used for locating threat
        # rasters, which are relative to this path
        threat_csv_dirpath = os.path.dirname(args['threats_table_path'])

        # Validate threat raster paths and their nodata values
        bad_threat_paths = []
        duplicate_paths = []
        threat_path_list = []
        bad_threat_columns = []
        for lulc_key, lulc_arg in (('_c', 'lulc_cur_path'),
                                   ('_f', 'lulc_fut_path'),
                                   ('_b', 'lulc_bas_path')):
            if lulc_arg in args and args[lulc_arg] != '':
                # for each threat given in the CSV file try opening the
                # associated raster which should be found in
                # threat_raster_folder
                for threat in threat_dict:
                    threat_table_path_col = _THREAT_SCENARIO_MAP[lulc_key]
                    if threat_table_path_col not in threat_dict[threat]:
                        bad_threat_columns.append(threat_table_path_col)
                        break

                    # Threat path from threat CSV is relative to CSV
                    threat_path = os.path.join(
                        threat_csv_dirpath,
                        threat_dict[threat][threat_table_path_col])

                    threat_validate_result = _validate_threat_path(
                        threat_path, lulc_key)
                    if threat_validate_result == 'error':
                        bad_threat_paths.append(
                            (threat, threat_table_path_col))
                        continue

                    threat_path = threat_validate_result

                    if threat_path:
                        # check for duplicate absolute threat path names that
                        # cause errors when trying to write aligned versions
                        if threat_path not in threat_path_list:
                            threat_path_list.append(threat_path)
                        else:
                            duplicate_paths.append(
                                os.path.basename(threat_path))

        if bad_threat_columns:
            validation_warnings.append((
                ['threats_table_path'],
                MISSING_COLUMN_MSG.format(column_name=bad_threat_columns[0])))

        if bad_threat_paths:
            validation_warnings.append((
                ['threats_table_path'],
                MISSING_THREAT_RASTER_MSG.format(threat_list=bad_threat_paths)
            ))
            invalid_keys.add('threats_table_path')

        if duplicate_paths:
            validation_warnings.append((
                ['threats_table_path'],
                DUPLICATE_PATHS_MSG + str(duplicate_paths)))

            if 'threats_table_path' not in invalid_keys:
                invalid_keys.add('threats_table_path')

    return validation_warnings



"""DelineateIt wrapper for pygeoprocessing's watershed delineation routine."""
import os
import logging
import time

from osgeo import gdal
from osgeo import ogr
from osgeo import osr
import shapely.errors
import shapely.geometry
import shapely.wkb
import numpy
import pygeoprocessing
import pygeoprocessing.routing
import taskgraph

from .. import utils
from .. import spec_utils
from ..spec_utils import u
from .. import validation
from .. import MODEL_METADATA
from . import delineateit_core


LOGGER = logging.getLogger(__name__)

ARGS_SPEC = {
    "model_name": MODEL_METADATA["delineateit"].model_title,
    "pyname": MODEL_METADATA["delineateit"].pyname,
    "userguide_html": MODEL_METADATA["delineateit"].userguide,
    "args_with_spatial_overlap": {
        "spatial_keys": ["dem_path", "outlet_vector_path"],
        "different_projections_ok": True,
    },
    "args": {
        "workspace_dir": spec_utils.WORKSPACE,
        "results_suffix": spec_utils.SUFFIX,
        "n_workers": spec_utils.N_WORKERS,
        "dem_path": {
            **spec_utils.DEM,
            "projected": True
        },
        "detect_pour_points": {
            "type": "boolean",
            "required": False,
            "about": _(
                "Detect pour points (watershed outlets) based on "
                "the DEM, and use these instead of a user-provided outlet "
                "features vector."),
            "name": _("detect pour points")
        },
        "outlet_vector_path": {
            "type": "vector",
            "fields": {},
            "geometries": spec_utils.ALL_GEOMS,
            "required": "not detect_pour_points",
            "about": _(
                "A map of watershed outlets from which to delineate the "
                "watersheds. Required if Detect Pour Points is not checked."),
            "name": _("watershed outlets")
        },
        "snap_points": {
            "type": "boolean",
            "required": False,
            "about": _(
                "Whether to snap point geometries to the nearest stream "
                "pixel.  If ``True``, ``args['flow_threshold']`` and "
                "``args['snap_distance']`` must also be defined. If a point "
                "is equally near to more than one stream pixel, it will be "
                "snapped to the stream pixel with the highest flow "
                "accumulation value. This has no effect if Detect Pour Points "
                "is selected."),
            "name": _("snap points to the nearest stream")
        },
        "flow_threshold": {
            **spec_utils.THRESHOLD_FLOW_ACCUMULATION,
            "required": "snap_points",
            "about": _(
                spec_utils.THRESHOLD_FLOW_ACCUMULATION["about"] +
                " Required if Snap Points is selected."),
        },
        "snap_distance": {
            "expression": "value > 0",
            "type": "number",
            "units": u.pixels,
            "required": "snap_points",
            "about": _(
                "Maximum distance to relocate watershed outlet points in "
                "order to snap them to a stream. Required if Snap Points "
                "is selected."),
            "name": _("snap distance")
        },
        "skip_invalid_geometry": {
            "type": "boolean",
            "required": False,
            "about": _(
                "Skip delineation for any invalid geometries found in the "
                "Outlet Features. Otherwise, an invalid geometry will cause "
                "the model to crash."),
            "name": _("skip invalid geometries")
        }
    }
}

_OUTPUT_FILES = {
    'preprocessed_geometries': 'preprocessed_geometries.gpkg',
    'filled_dem': 'filled_dem.tif',
    'flow_dir_d8': 'flow_direction.tif',
    'flow_accumulation': 'flow_accumulation.tif',
    'streams': 'streams.tif',
    'snapped_outlets': 'snapped_outlets.gpkg',
    'watersheds': 'watersheds.gpkg',
    'pour_points': 'pour_points.gpkg'
}


def execute(args):
    """DelineateIt: Watershed Delineation.

    This 'model' provides an InVEST-based wrapper around the pygeoprocessing
    routing API for watershed delineation.

    Upon successful completion, the following files are written to the
    output workspace:

        * ``snapped_outlets.gpkg`` - A GeoPackage with the points snapped
          to a nearby stream.
        * ``watersheds.gpkg`` - a GeoPackage of watersheds determined
          by the D8 routing algorithm.
        * ``stream.tif`` - a GeoTiff representing detected streams based on
          the provided ``flow_threshold`` parameter.  Values of 1 are
          streams, values of 0 are not.

    Args:
        args['workspace_dir'] (string):  The selected folder is used as the
            workspace all intermediate and output files will be written.If the
            selected folder does not exist, it will be created. If datasets
            already exist in the selected folder, they will be overwritten.
            (required)
        args['results_suffix'] (string):  This text will be appended to the end
            of output files to help separate multiple runs. (optional)
        args['dem_path'] (string):  A GDAL-supported raster file with an
            elevation for each cell. Make sure the DEM is corrected by filling
            in sinks, and if necessary burning hydrographic features into the
            elevation model (recommended when unusual streams are observed.)
            See the 'Working with the DEM' section of the InVEST User's Guide
            for more information. (required)
        args['outlet_vector_path'] (string):  This is a vector representing
            geometries that the watersheds should be built around. Required if
            ``args['detect_pour_points']`` is False; not used otherwise.
        args['snap_points'] (bool): Whether to snap point geometries to the
            nearest stream pixel.  If ``True``, ``args['flow_threshold']``
            and ``args['snap_distance']`` must also be defined.
        args['flow_threshold'] (int):  The number of upstream cells that must
            into a cell before it's considered part of a stream such that
            retention stops and the remaining export is exported to the stream.
            Used to define streams from the DEM.
        args['snap_distance'] (int):  Pixel Distance to Snap Outlet Points
        args['skip_invalid_geometry'] (bool): Whether to crash when an
            invalid geometry is passed or skip it, including all valid
            geometries in the vector to be passed to delineation.
            If ``False``, this tool will crash if an invalid geometry is
            found.  If ``True``, invalid geometries will be left out of
            the vector to be delineated.  Default: True
        args['detect_pour_points'] (bool): Whether to run the pour point
            detection algorithm. If True, detected pour points are used instead
            of outlet_vector_path geometries. Default: False
        args['n_workers'] (int): The number of worker processes to use with
            taskgraph. Defaults to -1 (no parallelism).

    Returns:
        ``None``

    """
    output_directory = args['workspace_dir']
    utils.make_directories([output_directory])

    file_suffix = utils.make_suffix_string(args, 'results_suffix')
    file_registry = utils.build_file_registry(
        [(_OUTPUT_FILES, output_directory)], file_suffix)

    work_token_dir = os.path.join(output_directory, '_work_tokens')

    # Manually setting n_workers to be -1 so that everything happens in the
    # same thread.
    try:
        n_workers = int(args['n_workers'])
    except (KeyError, TypeError, ValueError):
        # KeyError when n_workers is not present in args
        # ValueError when n_workers is an empty string.
        # TypeError when n_workers is None.
        n_workers = -1
    graph = taskgraph.TaskGraph(work_token_dir, n_workers=n_workers)

    fill_pits_task = graph.add_task(
        pygeoprocessing.routing.fill_pits,
        args=((args['dem_path'], 1),
              file_registry['filled_dem']),
        kwargs={'working_dir': output_directory},
        target_path_list=[file_registry['filled_dem']],
        task_name='fill_pits')

    flow_dir_task = graph.add_task(
        pygeoprocessing.routing.flow_dir_d8,
        args=((file_registry['filled_dem'], 1),
              file_registry['flow_dir_d8']),
        kwargs={'working_dir': output_directory},
        target_path_list=[file_registry['flow_dir_d8']],
        dependent_task_list=[fill_pits_task],
        task_name='flow_direction')

    if 'detect_pour_points' in args and args['detect_pour_points']:
        # Detect pour points automatically and use them instead of
        # user-provided geometries
        pour_points_task = graph.add_task(
            detect_pour_points,
            args=((file_registry['flow_dir_d8'], 1),
                  file_registry['pour_points']),
            dependent_task_list=[flow_dir_task],
            target_path_list=[file_registry['pour_points']],
            task_name='detect_pour_points')
        outlet_vector_path = file_registry['pour_points']
        geometry_task = pour_points_task
    else:
        check_geometries_task = graph.add_task(
            check_geometries,
            args=(args['outlet_vector_path'],
                  file_registry['filled_dem'],
                  file_registry['preprocessed_geometries'],
                  args.get('skip_invalid_geometry', True)),
            dependent_task_list=[fill_pits_task],
            target_path_list=[file_registry['preprocessed_geometries']],
            task_name='check_geometries')
        outlet_vector_path = file_registry['preprocessed_geometries']
        geometry_task = check_geometries_task

    delineation_dependent_tasks = [flow_dir_task, geometry_task]
    if 'snap_points' in args and args['snap_points']:
        flow_accumulation_task = graph.add_task(
            pygeoprocessing.routing.flow_accumulation_d8,
            args=((file_registry['flow_dir_d8'], 1),
                  file_registry['flow_accumulation']),
            target_path_list=[file_registry['flow_accumulation']],
            dependent_task_list=[flow_dir_task],
            task_name='flow_accumulation')
        delineation_dependent_tasks.append(flow_accumulation_task)

        snap_distance = int(args['snap_distance'])
        flow_threshold = int(args['flow_threshold'])

        out_nodata = 255
        flow_accumulation_task.join()  # wait so we can read the nodata value
        flow_accumulation_nodata = pygeoprocessing.get_raster_info(
            file_registry['flow_accumulation'])['nodata']
        streams_task = graph.add_task(
            pygeoprocessing.raster_calculator,
            args=([(file_registry['flow_accumulation'], 1),
                   (flow_accumulation_nodata, 'raw'),
                   (out_nodata, 'raw'),
                   (flow_threshold, 'raw')],
                  _threshold_streams,
                  file_registry['streams'],
                  gdal.GDT_Byte,
                  out_nodata),
            target_path_list=[file_registry['streams']],
            dependent_task_list=[flow_accumulation_task],
            task_name='threshold_streams')

        snapped_outflow_points_task = graph.add_task(
            snap_points_to_nearest_stream,
            args=(outlet_vector_path,
                  file_registry['streams'],
                  file_registry['flow_accumulation'],
                  snap_distance,
                  file_registry['snapped_outlets']),
            target_path_list=[file_registry['snapped_outlets']],
            dependent_task_list=[streams_task, geometry_task],
            task_name='snapped_outflow_points')
        delineation_dependent_tasks.append(snapped_outflow_points_task)
        outlet_vector_path = file_registry['snapped_outlets']

    _ = graph.add_task(
        pygeoprocessing.routing.delineate_watersheds_d8,
        args=((file_registry['flow_dir_d8'], 1),
              outlet_vector_path,
              file_registry['watersheds']),
        kwargs={'working_dir': output_directory,
                'target_layer_name':
                    os.path.splitext(
                        os.path.basename(file_registry['watersheds']))[0]},
        target_path_list=[file_registry['watersheds']],
        dependent_task_list=delineation_dependent_tasks,
        task_name='delineate_watersheds_single_worker')

    graph.close()
    graph.join()


def _threshold_streams(flow_accum, src_nodata, out_nodata, threshold):
    """Identify stream pixels based on a user-defined threshold.

    This is an ``op`` to ``pygeoprocessing.raster_calculator``.  Any non-nodata
    pixels in ``flow_accum`` greater than the value of ``threshold`` are
    marked as stream pixels.  Any non-nodata pixels below ``threshold`` are
    marked as non-stream pixels.

    Args:
        flow_accum (numpy array): A numpy array of flow accumulation values.
        src_nodata (number): A number indicating the nodata value of the
            flow accumulation array.
        out_nodata (number): A number indicating the nodata value of the
            target array.
        threshold (number): A numeric threshold over which a flow
            accumulation pixel will be marked as a stream.

    Returns:
        A ``numpy.uint8`` array with values of 0, 1 or ``out_nodata``.

    """
    out_matrix = numpy.empty(flow_accum.shape, dtype=numpy.uint8)
    out_matrix[:] = out_nodata

    valid_pixels = slice(None)
    if src_nodata is not None:
        valid_pixels = ~utils.array_equals_nodata(flow_accum, src_nodata)

    over_threshold = flow_accum > threshold
    out_matrix[valid_pixels & over_threshold] = 1
    out_matrix[valid_pixels & ~over_threshold] = 0
    return out_matrix


def check_geometries(outlet_vector_path, dem_path, target_vector_path,
                     skip_invalid_geometry=False):
    """Perform reasonable checks and repairs on the incoming vector.

    This function will iterate through the vector at ``outlet_vector_path``
    and validate geometries, putting the geometries into a new geopackage
    at ``target_vector_path``.

    The vector at ``target_vector_path`` will include features that:

        * Have valid geometries
        * Are simplified to 1/2 the DEM pixel size
        * Intersect the bounding box of the DEM

    Any geometries that are empty or do not intersect the DEM will not be
    included in ``target_vector_path``.

    Args:
        outlet_vector_path (string): The path to an outflow vector.  The first
            layer of the vector only will be inspected.
        dem_path (string): The path to a DEM on disk.
        target_vector_path (string): The target path to where the output
            geopackage should be written.
        skip_invalid_geometry (bool): Whether to raise an exception
            when invalid geometry is found.  If ``False``, an exception
            will be raised when the first invalid geometry is found.
            If ``True``, the invalid geometry will be not be included
            in the output vector but any other valid geometries will.

    Returns:
        ``None``

    """
    if os.path.exists(target_vector_path):
        LOGGER.debug('Target vector path %s exists on disk; removing.',
                     target_vector_path)
        os.remove(target_vector_path)

    dem_info = pygeoprocessing.get_raster_info(dem_path)
    dem_bbox = shapely.prepared.prep(
        shapely.geometry.box(*dem_info['bounding_box']))
    nyquist_limit = numpy.mean(numpy.abs(dem_info['pixel_size'])) / 2.

    dem_srs = osr.SpatialReference()
    dem_srs.ImportFromWkt(dem_info['projection_wkt'])

    gpkg_driver = gdal.GetDriverByName('GPKG')
    target_vector = gpkg_driver.Create(target_vector_path, 0, 0, 0,
                                       gdal.GDT_Unknown)
    layer_name = os.path.splitext(os.path.basename(target_vector_path))[0]
    target_layer = target_vector.CreateLayer(
        layer_name, dem_srs, ogr.wkbUnknown)  # Use source layer type?

    outflow_vector = gdal.OpenEx(outlet_vector_path, gdal.OF_VECTOR)
    outflow_layer = outflow_vector.GetLayer()
    LOGGER.info('Checking %s geometries from source vector',
                outflow_layer.GetFeatureCount())
    target_layer.CreateFields(outflow_layer.schema)

    target_layer.StartTransaction()
    for feature in outflow_layer:
        original_geometry = feature.GetGeometryRef()

        try:
            shapely_geom = shapely.wkb.loads(
                bytes(original_geometry.ExportToWkb()))

            # The classic bowtie polygons will load but require a separate
            # check for validity.
            if not shapely_geom.is_valid:
                raise ValueError('Shapely geom is invalid.')
        except (shapely.errors.ReadingError, ValueError):
            # Parent class for shapely GEOS errors
            # Raised when the geometry is invalid.
            if not skip_invalid_geometry:
                outflow_layer = None
                outflow_vector = None
                target_layer = None
                target_vector = None
                raise ValueError(
                    "The geometry at feature %s is invalid.  Check the logs "
                    "for details and try re-running with repaired geometry."
                    % feature.GetFID())
            else:
                LOGGER.warning(
                    "The geometry at feature %s is invalid and will not be "
                    "included in the set of features to be delineated.",
                    feature.GetFID())
                continue

        if shapely_geom.is_empty:
            LOGGER.warning(
                'Feature %s has no geometry. Skipping', feature.GetFID())
            continue

        shapely_bbox = shapely.geometry.box(*shapely_geom.bounds)
        if not dem_bbox.intersects(shapely_bbox):
            LOGGER.warning('Feature %s does not intersect the DEM. Skipping.',
                           feature.GetFID())
            continue

        simplified_geometry = shapely_geom.simplify(nyquist_limit)

        new_feature = ogr.Feature(target_layer.GetLayerDefn())
        new_feature.SetGeometry(ogr.CreateGeometryFromWkb(
            simplified_geometry.wkb))
        for field_name, field_value in feature.items().items():
            new_feature.SetField(field_name, field_value)
        target_layer.CreateFeature(new_feature)

    target_layer.CommitTransaction()

    LOGGER.info('%s features copied to %s from the original %s features',
                target_layer.GetFeatureCount(),
                os.path.basename(target_vector_path),
                outflow_layer.GetFeatureCount())
    outflow_layer = None
    outflow_vector = None
    target_layer = None
    target_vector = None


def snap_points_to_nearest_stream(points_vector_path, stream_raster_path,
                                  flow_accum_raster_path, snap_distance,
                                  snapped_points_vector_path):
    """Adjust the location of points to the nearest stream pixel.

    The new point layer will have all fields and field values copied over from
    the source vector.  Any points that are outside of the stream raster will
    not be included in the output vector.

    Args:
        points_vector_path (string): A path to a vector on disk containing
            point geometries.  Must be in the same projection as the stream
            raster.
        stream_raster_path (string): A path to a stream raster, where
            pixel values are ``1`` (indicating a stream pixel) or ``0``
            (indicating a non-stream pixel).
        flow_accum_raster_path (string): A path to a flow accumulation raster
            that is aligned with the stream raster. Used to break ties between
            equally-near stream pixels.
        snap_distance (number): The maximum distance (in pixels) to search
            for stream pixels for each point.  This must be a positive, nonzero
            value.
        snapped_points_vector_path (string): A path to where the output
            points will be written.

    Returns:
        ``None``

    Raises:
        ``ValueError`` when snap_distance is less than or equal to 0.

    """
    if snap_distance <= 0:
        raise ValueError('Snap_distance must be >= 0, not %s' % snap_distance)

    points_vector = gdal.OpenEx(points_vector_path, gdal.OF_VECTOR)
    points_layer = points_vector.GetLayer()

    stream_raster_info = pygeoprocessing.get_raster_info(stream_raster_path)
    geotransform = stream_raster_info['geotransform']
    n_cols, n_rows = stream_raster_info['raster_size']
    stream_raster = gdal.OpenEx(stream_raster_path, gdal.OF_RASTER)
    stream_band = stream_raster.GetRasterBand(1)

    flow_accum_raster = gdal.OpenEx(flow_accum_raster_path, gdal.OF_RASTER)
    flow_accum_band = flow_accum_raster.GetRasterBand(1)

    driver = gdal.GetDriverByName('GPKG')
    snapped_vector = driver.Create(snapped_points_vector_path, 0, 0, 0,
                                   gdal.GDT_Unknown)
    layer_name = os.path.splitext(
        os.path.basename(snapped_points_vector_path))[0]
    snapped_layer = snapped_vector.CreateLayer(
        layer_name, points_layer.GetSpatialRef(), points_layer.GetGeomType())
    snapped_layer.CreateFields(points_layer.schema)
    snapped_layer_defn = snapped_layer.GetLayerDefn()

    snapped_layer.StartTransaction()
    n_features = points_layer.GetFeatureCount()
    last_time = time.time()
    for index, point_feature in enumerate(points_layer, 1):
        if time.time() - last_time > 5.0:
            LOGGER.info('Snapped %s of %s points', index, n_features)
            last_time = time.time()

        source_geometry = point_feature.GetGeometryRef()
        geom_name = source_geometry.GetGeometryName()
        geom_count = source_geometry.GetGeometryCount()

        if source_geometry.IsEmpty():
            LOGGER.warning(
                f"FID {point_feature.GetFID()} is missing a defined geometry. "
                "Skipping.")
            continue

        # If the geometry is not a primitive point, just create the new feature
        # as it is now in the new vector. MULTIPOINT geometries with a single
        # component point count as primitive points.
        # OGR's wkbMultiPoint, wkbMultiPointM, wkbMultiPointZM and
        # wkbMultiPoint25D all use the MULTIPOINT geometry name.
        if ((geom_name not in ('POINT', 'MULTIPOINT')) or
                (geom_name == 'MULTIPOINT' and geom_count > 1)):
            LOGGER.warning(
                f"FID {point_feature.GetFID()} ({geom_name}, n={geom_count}) "
                "Geometry cannot be snapped.")
            new_feature = ogr.Feature(snapped_layer.GetLayerDefn())
            new_feature.SetGeometry(source_geometry)
            for field_name, field_value in point_feature.items().items():
                new_feature.SetField(field_name, field_value)
            snapped_layer.CreateFeature(new_feature)
            continue

        point = shapely.wkb.loads(bytes(source_geometry.ExportToWkb()))
        if geom_name == 'MULTIPOINT':
            # We already checked (above) that there's only one component point
            point = point.geoms[0]

        x_index = (point.x - geotransform[0]) // geotransform[1]
        y_index = (point.y - geotransform[3]) // geotransform[5]
        if (x_index < 0 or x_index > n_cols or
                y_index < 0 or y_index > n_rows):
            LOGGER.warning(
                'Encountered a point that was outside the bounds of the '
                f'stream raster.  FID:{point_feature.GetFID()} at {point}')
            continue

        x_left = max(x_index - snap_distance, 0)
        y_top = max(y_index - snap_distance, 0)
        x_right = min(x_index + snap_distance, n_cols)
        y_bottom = min(y_index + snap_distance, n_rows)

        # snap to the nearest stream pixel out to the snap distance
        stream_window = stream_band.ReadAsArray(
            int(x_left), int(y_top), int(x_right - x_left),
            int(y_bottom - y_top))
        row_indexes, col_indexes = numpy.nonzero(
            stream_window == 1)

        # Find the closest stream pixel that meets the distance
        # requirement. If there is a tie, snap to the stream pixel with
        # a higher flow accumulation value.
        if row_indexes.size > 0:  # there are streams within the snap distance
            # Calculate euclidean distance from the point to each stream pixel
            distance_array = numpy.hypot(
                # distance along y axis from the point to each stream pixel
                y_index - y_top - row_indexes,
                # distance along x axis from the point to each stream pixel
                x_index - x_left - col_indexes,
                dtype=numpy.float32)

            is_nearest = distance_array == distance_array.min()
            # if > 1 stream pixel is nearest, break tie with flow accumulation
            if is_nearest.sum() > 1:
                flow_accum_array = flow_accum_band.ReadAsArray(
                    int(x_left), int(y_top), int(x_right - x_left),
                    int(y_bottom - y_top))
                # weight by flow accum
                is_nearest = is_nearest * flow_accum_array[row_indexes, col_indexes]

            # 1d index of max value in flattened array
            nearest_stream_index_1d = numpy.argmax(is_nearest)

            # convert 1d index back to coordinates relative to window
            nearest_stream_row = row_indexes[nearest_stream_index_1d]
            nearest_stream_col = col_indexes[nearest_stream_index_1d]

            offset_row = nearest_stream_row - (y_index - y_top)
            offset_col = nearest_stream_col - (x_index - x_left)

            y_index += offset_row
            x_index += offset_col

        point_geometry = ogr.Geometry(ogr.wkbPoint)
        point_geometry.AddPoint(
            geotransform[0] + (x_index + 0.5) * geotransform[1],
            geotransform[3] + (y_index + 0.5) * geotransform[5])

        # Get the output Layer's Feature Definition
        snapped_point_feature = ogr.Feature(snapped_layer_defn)
        for field_name, field_value in point_feature.items().items():
            snapped_point_feature.SetField(field_name, field_value)
        snapped_point_feature.SetGeometry(point_geometry)

        snapped_layer.CreateFeature(snapped_point_feature)
    snapped_layer.CommitTransaction()
    snapped_layer = None
    snapped_vector = None

    points_layer = None
    points_vector = None


def detect_pour_points(flow_dir_raster_path_band, target_vector_path):
    """
    Create a pour point vector from D8 flow direction raster.

    A pour point is the center point of a pixel which:
        - flows off of the raster, or
        - flows into a nodata pixel

    Args:
        flow_dir_raster_path_band (tuple): tuple of (raster path, band index)
            indicating the flow direction raster to use. Pixel values are D8
            values [0 - 7] in this order:

                321
                4x0
                567

        target_vector_path (string): path to save pour point vector to.

    Returns:
        None
    """
    raster_info = pygeoprocessing.get_raster_info(flow_dir_raster_path_band[0])
    pour_point_set = _find_raster_pour_points(flow_dir_raster_path_band)

    # use same spatial reference as the input
    aoi_spatial_reference = osr.SpatialReference()
    aoi_spatial_reference.ImportFromWkt(raster_info['projection_wkt'])

    gpkg_driver = ogr.GetDriverByName("GPKG")
    target_vector = gpkg_driver.CreateDataSource(target_vector_path)
    layer_name = os.path.splitext(
        os.path.basename(target_vector_path))[0]
    target_layer = target_vector.CreateLayer(
        layer_name, aoi_spatial_reference, ogr.wkbPoint)
    target_defn = target_layer.GetLayerDefn()

    # It's important to have a user-facing unique ID field for post-processing
    # (e.g. table-joins) that is not the FID. FIDs are not stable across file
    # conversions that users might do.
    target_layer.CreateField(
        ogr.FieldDefn('point_id', ogr.OFTInteger64))

    # Add a feature to the layer for each point
    target_layer.StartTransaction()
    for idx, (x, y) in enumerate(pour_point_set):
        geometry = ogr.Geometry(ogr.wkbPoint)
        geometry.AddPoint(x, y)
        feature = ogr.Feature(target_defn)
        feature.SetGeometry(geometry)
        feature.SetField('point_id', idx)
        target_layer.CreateFeature(feature)
    target_layer.CommitTransaction()

    target_layer = None
    target_vector = None


def _find_raster_pour_points(flow_dir_raster_path_band):
    """
    Memory-safe pour point calculation from a flow direction raster.

    Args:
        flow_dir_raster_path_band (tuple): tuple of (raster path, band index)
            indicating the flow direction raster to use.

    Returns:
        set of (x, y) coordinate tuples of pour points, in the same coordinate
        system as the input raster.
    """
    flow_dir_raster_path, band_index = flow_dir_raster_path_band
    raster_info = pygeoprocessing.get_raster_info(flow_dir_raster_path)
    # Open the flow direction raster band
    raster = gdal.OpenEx(flow_dir_raster_path, gdal.OF_RASTER)
    band = raster.GetRasterBand(band_index)
    width, height = raster_info['raster_size']

    pour_points = set()
    # Read in flow direction data and find pour points one block at a time
    for offsets in pygeoprocessing.iterblocks(
            (flow_dir_raster_path, band_index), offset_only=True):
        # Expand each block by a one-pixel-wide margin, if possible.
        # This way the blocks will overlap so the watershed
        # calculation will be continuous.
        if offsets['xoff'] > 0:
            offsets['xoff'] -= 1
            offsets['win_xsize'] += 1
        if offsets['yoff'] > 0:
            offsets['yoff'] -= 1
            offsets['win_ysize'] += 1
        if offsets['xoff'] + offsets['win_xsize'] < width:
            offsets['win_xsize'] += 1
        if offsets['yoff'] + offsets['win_ysize'] < height:
            offsets['win_ysize'] += 1

        # Keep track of which block edges are raster edges
        edges = numpy.empty(4, dtype=numpy.intc)
        # edges order: top, left, bottom, right
        edges[0] = (offsets['yoff'] == 0)
        edges[1] = (offsets['xoff'] == 0)
        edges[2] = (offsets['yoff'] + offsets['win_ysize'] == height)
        edges[3] = (offsets['xoff'] + offsets['win_xsize'] == width)

        flow_dir_block = band.ReadAsArray(**offsets)
        pour_points = pour_points.union(
            delineateit_core.calculate_pour_point_array(
                # numpy.intc is equivalent to an int in C (normally int32 or
                # int64). This way it can be passed directly into a memoryview
                # (int[:, :]) in the Cython function.
                flow_dir_block.astype(numpy.intc),
                edges,
                nodata=raster_info['nodata'][band_index - 1],
                offset=(offsets['xoff'], offsets['yoff']),
                origin=(raster_info['geotransform'][0],
                        raster_info['geotransform'][3]),
                pixel_size=raster_info['pixel_size']))

    raster = None
    band = None

    return pour_points


@validation.invest_validator
def validate(args, limit_to=None):
    """Validate args to ensure they conform to `execute`'s contract.

    Args:
        args (dict): dictionary of key(str)/value pairs where keys and
            values are specified in `execute` docstring.
        limit_to (str): (optional) if not None indicates that validation
            should only occur on the args[limit_to] value. The intent that
            individual key validation could be significantly less expensive
            than validating the entire `args` dictionary.

    Returns:
        list of ([invalid key_a, invalid_keyb, ...], 'warning/error message')
            tuples. Where an entry indicates that the invalid keys caused
            the error message in the second part of the tuple. This should
            be an empty list if validation succeeds.

    """
    return validation.validate(
        args, ARGS_SPEC['args'], ARGS_SPEC['args_with_spatial_overlap'])
